from __future__ import annotations

from collections.abc import Mapping, Sequence
from collections.abc import Set as AbstractSet
from pathlib import Path
from typing import Any, cast

from fastparquet import write
from hypothesis import assume, given
from hypothesis.extra.pandas import column, data_frames, range_indexes
from hypothesis.strategies import (
    DataObject,
    SearchStrategy,
    booleans,
    data,
    dictionaries,
    floats,
    integers,
    lists,
    none,
    sampled_from,
    sets,
    tuples,
)
from numpy import float32, nan
from pandas import DataFrame, RangeIndex, Series, concat
from pandas.testing import assert_frame_equal, assert_series_equal
from pytest import mark, param, raises

from utilities.fastparquet import (
    _PARQUET_DTYPES,
    GetParquetFileError,
    WriteParquetCoreError,
    _get_parquet_file,
    count_rows,
    get_columns,
    get_dtypes,
    get_num_row_groups,
    read_parquet,
    write_parquet,
)
from utilities.hypothesis import dates_pd, lists_fixed_length, temp_paths, text_ascii
from utilities.numpy import dt64ns
from utilities.pandas import Int64, astype, string
from utilities.pytest import skipif_windows


class TestCountRows:
    @given(rows=lists(booleans(), min_size=1), root=temp_paths())
    def test_main(self, *, rows: Sequence[bool], root: Path) -> None:
        n = len(rows)
        df = DataFrame(rows, index=RangeIndex(n), columns=["value"])
        write_parquet(df, path := root.joinpath("df.parq"))
        result = count_rows(path)
        assert result == n


class TestGetColumns:
    @given(columns=sets(text_ascii()), root=temp_paths())
    def test_main(self, *, columns: AbstractSet[str], root: Path) -> None:
        as_list = list(columns)
        df = DataFrame(nan, index=RangeIndex(1), columns=as_list, dtype=float)
        write_parquet(df, path := root.joinpath("df.parq"))
        result = get_columns(path)
        assert result == as_list


class TestGetDtypes:
    @given(
        dtypes=dictionaries(
            text_ascii(), sampled_from(sorted(_PARQUET_DTYPES, key=repr))
        ),
        root=temp_paths(),
    )
    def test_main(self, *, dtypes: Mapping[str, Any], root: Path) -> None:
        df = DataFrame(None, index=RangeIndex(1), columns=list(dtypes))
        df = astype(df, dtypes)
        write_parquet(df, path := root.joinpath("df.parq"))
        result = get_dtypes(path)
        assert result == dtypes


class TestGetNumRowGroups:
    def test_main(self, *, tmp_path: Path) -> None:
        df = DataFrame(0.0, index=RangeIndex(2), columns=["value"])
        write_parquet(df, path := tmp_path.joinpath("df.parq"), row_group_offsets=1)
        result = get_num_row_groups(path)
        expected = 2
        assert result == expected


class TestGetParquetFile:
    def test_main(self, *, tmp_path: Path) -> None:
        path = tmp_path.joinpath("file")
        df = DataFrame(0.0, index=RangeIndex(1), columns=["value"])
        write(path, df)
        _ = _get_parquet_file(path)

    def test_get_row_group(self, *, tmp_path: Path) -> None:
        path = tmp_path.joinpath("file")
        df = DataFrame(0.0, index=RangeIndex(2), columns=["value"])
        write(path, df, row_group_offsets=1)
        _ = _get_parquet_file(path, row_group=0)
        _ = _get_parquet_file(path, row_group=1)
        with raises(GetParquetFileError):
            _ = _get_parquet_file(path, row_group=2)

    @mark.parametrize("as_str", [param(True), param(False, marks=skipif_windows)])
    def test_error(self, *, tmp_path: Path, as_str: bool) -> None:
        path = tmp_path.joinpath("file")
        path_use = str(path) if as_str else path
        with raises(FileNotFoundError):
            _ = _get_parquet_file(path_use)


class TestReadAndWriteParquet:
    @given(data=data(), root=temp_paths(), as_series=booleans())
    @mark.parametrize(
        ("elements", "dtype"),
        [
            param(booleans(), bool),
            param(dates_pd() | none(), dt64ns),
            param(floats(-10.0, 10.0) | none(), float),
            param(integers(-10, 10) | none(), Int64),
            param(text_ascii() | none(), string),
        ],
    )
    def test_writing_df(
        self,
        *,
        data: DataObject,
        elements: SearchStrategy[Any],
        dtype: Any,
        root: Path,
        as_series: bool,
    ) -> None:
        rows = data.draw(lists(elements, min_size=1))
        n = len(rows)
        df = DataFrame(rows, index=RangeIndex(n), columns=["value"])
        df = astype(df, dtype)
        write_parquet(df, path := root.joinpath("df.parq"))
        head = data.draw(sampled_from([n, None]))
        columns = "value" if as_series else None
        read = read_parquet(path, head=head, columns=columns)
        if as_series:
            assert isinstance(read, Series)
            assert_series_equal(read, df["value"])
        else:
            assert isinstance(read, DataFrame)
            assert_frame_equal(read, df)

    @given(value1=floats(), value2=floats(), root=temp_paths())
    def test_writing_iterable_of_dfs(
        self, *, value1: float, value2: float, root: Path
    ) -> None:
        df = DataFrame([[value1], [value2]], index=RangeIndex(2), columns=["value"])
        parts = [df.iloc[:1], df.iloc[1:].reset_index(drop=True)]
        write_parquet(parts, path := root.joinpath("df.parq"))
        result = read_parquet(path)
        assert_frame_equal(result, df)

    @given(data=data(), column1=text_ascii(), column2=text_ascii(), root=temp_paths())
    def test_series_from_dataframe_with_two_string_columns(
        self, *, data: DataObject, column1: str, column2: str, root: Path
    ) -> None:
        _ = assume(column1 != column2)
        elements = text_ascii() | none()
        rows = data.draw(lists(tuples(elements, elements), min_size=1))
        df = DataFrame(
            rows, index=RangeIndex(len(rows)), columns=[column1, column2], dtype=string
        )
        write_parquet(df, path := root.joinpath("df.parq"))
        sr = read_parquet(path, columns=column1)
        assert_series_equal(sr, df[column1])

    @given(value1=floats(), value2=floats(), root=temp_paths())
    def test_read_row_groups(self, *, value1: float, value2: float, root: Path) -> None:
        df = DataFrame([[value1], [value2]], index=RangeIndex(2), columns=["value"])
        write_parquet(df, path := root.joinpath("df.parq"), row_group_offsets=1)
        result1 = read_parquet(path, row_group=0)
        expected1 = df.iloc[:1]
        assert_frame_equal(result1, expected1)
        result2 = read_parquet(path, row_group=1)
        expected2 = df.iloc[1:].reset_index(drop=True)
        assert_frame_equal(result2, expected2)
        with raises(GetParquetFileError):
            _ = read_parquet(path, row_group=2)

    @given(
        data=data(),
        root=temp_paths(),
        num_dfs=integers(1, 10),
        row_group_offsets=integers(1, 10),
    )
    def test_iterable_of_dfs_with_strings(
        self, *, data: DataObject, root: Path, num_dfs: int, row_group_offsets: int
    ) -> None:
        def as_str(df: DataFrame, /) -> DataFrame:
            return astype(df, string)

        elements = data_frames(
            [cast(Any, column)("value", elements=text_ascii())],
            index=range_indexes(min_size=1),
        ).map(as_str)
        dfs = data.draw(lists_fixed_length(elements, num_dfs))
        write_parquet(
            dfs, path := root.joinpath("df.parq"), row_group_offsets=row_group_offsets
        )
        result = read_parquet(path)
        expected = concat(dfs).reset_index(drop=True)
        assert_frame_equal(result, expected)


class TestWriteParquetCore:
    @given(value=text_ascii(), root=temp_paths())
    def test_strings_are_stored_as_objects(self, *, value: str, root: Path) -> None:
        df = DataFrame(value, index=RangeIndex(1), columns=["value"], dtype=string)
        write_parquet(df, path := root.joinpath("df.parq"))
        file = _get_parquet_file(path)
        dtypes = file.dtypes
        expected = {"value": object}
        assert dtypes == expected

    def test_extra_dtype(self, *, tmp_path: Path) -> None:
        df = DataFrame(nan, index=RangeIndex(1), columns=["value"], dtype=float32)
        write_parquet(df, tmp_path.joinpath("df.parq"), extra_dtypes={"value": float32})

    @mark.parametrize(
        "df",
        [
            param(DataFrame()),
            param(
                DataFrame(nan, index=RangeIndex(1), columns=["value"], dtype=float32)
            ),
        ],
    )
    def test_error(self, *, df: DataFrame, tmp_path: Path) -> None:
        with raises(WriteParquetCoreError):
            write_parquet(df, tmp_path.joinpath("df.parq"))

    def test_extra_dtype_ignored(self, *, tmp_path: Path) -> None:
        df = DataFrame(nan, index=RangeIndex(1), columns=["value"], dtype=float32)
        with raises(WriteParquetCoreError):
            write_parquet(
                df, tmp_path.joinpath("df.parq"), extra_dtypes={"other": float32}
            )



from __future__ import annotations

from collections.abc import Callable, Iterable, Mapping
from contextlib import suppress
from functools import reduce
from math import isclose
from operator import ge, itemgetter, le
from typing import Any, NoReturn, cast

import sqlalchemy
from sqlalchemy import (
    URL,
    Boolean,
    Column,
    Connection,
    DateTime,
    Engine,
    Float,
    Interval,
    LargeBinary,
    MetaData,
    Numeric,
    String,
    Table,
    Unicode,
    UnicodeText,
    Uuid,
    and_,
    case,
    quoted_name,
    text,
)
from sqlalchemy import create_engine as _create_engine
from sqlalchemy.exc import (
    ArgumentError,
    DatabaseError,
    NoSuchTableError,
    OperationalError,
)
from sqlalchemy.orm import declared_attr
from sqlalchemy.pool import NullPool, Pool
from sqlalchemy.sql.base import ReadOnlyColumnCollection
from typing_extensions import assert_never

from utilities._sqlalchemy.common import (
    INSERT_ITEMS_CHUNK_SIZE_FRAC,
    CheckSeriesAgainstTableColumnError,
    CheckSeriesAgainstTableSchemaError,
    Dialect,
    GetDialectError,
    GetTableError,
    InsertItemsCollectError,
    InsertItemsCollectIterableError,
    check_dataframe_schema_against_table,
    check_selectable_for_duplicate_columns,
    check_series_against_table_column,
    check_series_against_table_schema,
    get_column_names,
    get_columns,
    get_dialect,
    get_table,
    insert_items,
    insert_items_collect,
    insert_items_collect_iterable,
    is_mapped_class,
    is_table_or_mapped_class,
    mapped_class_to_dict,
    yield_connection,
)
from utilities.class_name import get_class_name
from utilities.errors import redirect_error
from utilities.math import FloatNonNeg, IntNonNeg
from utilities.text import ensure_str, snake_case, snake_case_mappings
from utilities.types import IterableStrs


def check_column_collections_equal(
    x: ReadOnlyColumnCollection[Any, Any],
    y: ReadOnlyColumnCollection[Any, Any],
    /,
    *,
    snake: bool = False,
    allow_permutations: bool = False,
    primary_key: bool = True,
) -> None:
    """Check that a pair of column collections are equal."""
    cols_x, cols_y = (list(cast(Iterable[Column[Any]], i)) for i in [x, y])
    name_to_col_x, name_to_col_y = (
        {ensure_str(col.name): col for col in i} for i in [cols_x, cols_y]
    )
    if len(name_to_col_x) != len(name_to_col_y):
        msg = f"{x=}, {y=}"
        raise CheckColumnCollectionsEqualError(msg)
    if snake:
        snake_to_name_x, snake_to_name_y = (
            snake_case_mappings(i, inverse=True) for i in [name_to_col_x, name_to_col_y]
        )
        key_to_col_x, key_to_col_y = (
            {key: name_to_col[snake_to_name[key]] for key in snake_to_name}
            for name_to_col, snake_to_name in [
                (name_to_col_x, snake_to_name_x),
                (name_to_col_y, snake_to_name_y),
            ]
        )
    else:
        key_to_col_x, key_to_col_y = name_to_col_x, name_to_col_y
    if allow_permutations:
        cols_to_check_x, cols_to_check_y = (
            map(itemgetter(1), sorted(key_to_col.items(), key=itemgetter(0)))
            for key_to_col in [key_to_col_x, key_to_col_y]
        )
    else:
        cols_to_check_x, cols_to_check_y = (
            i.values() for i in [key_to_col_x, key_to_col_y]
        )
    diff = set(key_to_col_x).symmetric_difference(set(key_to_col_y))
    if len(diff) >= 1:
        msg = f"{x=}, {y=}"
        raise CheckColumnCollectionsEqualError(msg)
    for x_i, y_i in zip(cols_to_check_x, cols_to_check_y, strict=True):
        check_columns_equal(x_i, y_i, snake=snake, primary_key=primary_key)


class CheckColumnCollectionsEqualError(Exception):
    ...


def check_columns_equal(
    x: Column[Any], y: Column[Any], /, *, snake: bool = False, primary_key: bool = True
) -> None:
    """Check that a pair of columns are equal."""
    check_table_or_column_names_equal(x.name, y.name, snake=snake)
    check_column_types_equal(x.type, y.type)
    if primary_key and (x.primary_key != y.primary_key):
        msg = f"{x.primary_key=}, {y.primary_key=}"
        raise CheckColumnsEqualError(msg)
    if x.nullable != y.nullable:
        msg = f"{x.nullable=}, {y.nullable=}"
        raise CheckColumnsEqualError(msg)


class CheckColumnsEqualError(Exception):
    ...


def check_column_types_equal(x: Any, y: Any, /) -> None:
    """Check that a pair of column types are equal."""
    x_inst, y_inst = (i() if isinstance(i, type) else i for i in [x, y])
    x_cls, y_cls = (i._type_affinity for i in [x_inst, y_inst])  # noqa: SLF001
    msg = f"{x=}, {y=}"
    if not (isinstance(x_inst, y_cls) and isinstance(y_inst, x_cls)):
        raise CheckColumnTypesEqualError(msg)
    if isinstance(x_inst, Boolean) and isinstance(y_inst, Boolean):
        check_column_types_boolean_equal(x_inst, y_inst)
    if isinstance(x_inst, DateTime) and isinstance(y_inst, DateTime):
        check_column_types_datetime_equal(x_inst, y_inst)
    if isinstance(x_inst, sqlalchemy.Enum) and isinstance(y_inst, sqlalchemy.Enum):
        check_column_types_enum_equal(x_inst, y_inst)
    if isinstance(x_inst, Float) and isinstance(y_inst, Float):
        check_column_types_float_equal(x_inst, y_inst)
    if isinstance(x_inst, Interval) and isinstance(y_inst, Interval):
        check_column_types_interval_equal(x_inst, y_inst)
    if isinstance(x_inst, LargeBinary) and isinstance(y_inst, LargeBinary):
        check_column_types_large_binary_equal(x_inst, y_inst)
    if isinstance(x_inst, Numeric) and isinstance(y_inst, Numeric):
        check_column_types_numeric_equal(x_inst, y_inst)
    if isinstance(x_inst, String | Unicode | UnicodeText) and isinstance(
        y_inst, String | Unicode | UnicodeText
    ):
        check_column_types_string_equal(x_inst, y_inst)
    if isinstance(x_inst, Uuid) and isinstance(y_inst, Uuid):
        check_column_types_uuid_equal(x_inst, y_inst)


class CheckColumnTypesEqualError(Exception):
    ...


def check_column_types_boolean_equal(x: Any, y: Any, /) -> None:
    """Check that a pair of boolean column types are equal."""
    msg = f"{x=}, {y=}"
    if x.create_constraint is not y.create_constraint:
        raise CheckColumnTypesBooleanEqualError(msg)
    if x.name != y.name:
        raise CheckColumnTypesBooleanEqualError(msg)


class CheckColumnTypesBooleanEqualError(Exception):
    ...


def check_column_types_datetime_equal(x: Any, y: Any, /) -> None:
    """Check that a pair of datetime column types are equal."""
    if x.timezone is not y.timezone:
        msg = f"{x=}, {y=}"
        raise CheckColumnTypesDateTimeEqualError(msg)


class CheckColumnTypesDateTimeEqualError(Exception):
    ...


def check_column_types_enum_equal(x: Any, y: Any, /) -> None:
    """Check that a pair of enum column types are equal."""
    x_enum, y_enum = (i.enum_class for i in [x, y])
    if (x_enum is None) and (y_enum is None):
        return
    msg = f"{x=}, {y=}"
    if ((x_enum is None) and (y_enum is not None)) or (
        (x_enum is not None) and (y_enum is None)
    ):
        raise CheckColumnTypesEnumEqualError(msg)
    if not (issubclass(x_enum, y_enum) and issubclass(y_enum, x_enum)):
        raise CheckColumnTypesEnumEqualError(msg)
    if x.create_constraint is not y.create_constraint:
        raise CheckColumnTypesEnumEqualError(msg)
    if x.native_enum is not y.native_enum:
        raise CheckColumnTypesEnumEqualError(msg)
    if x.length != y.length:
        raise CheckColumnTypesEnumEqualError(msg)
    if x.inherit_schema is not y.inherit_schema:
        raise CheckColumnTypesEnumEqualError(msg)


class CheckColumnTypesEnumEqualError(Exception):
    ...


def check_column_types_float_equal(x: Any, y: Any, /) -> None:
    """Check that a pair of float column types are equal."""
    msg = f"{x=}, {y=}"
    if x.precision != y.precision:
        raise CheckColumnTypesFloatEqualError(msg)
    if x.asdecimal is not y.asdecimal:
        raise CheckColumnTypesFloatEqualError(msg)
    if x.decimal_return_scale != y.decimal_return_scale:
        raise CheckColumnTypesFloatEqualError(msg)


class CheckColumnTypesFloatEqualError(Exception):
    ...


def check_column_types_interval_equal(x: Any, y: Any, /) -> None:
    """Check that a pair of interval column types are equal."""
    msg = f"{x=}, {y=}"
    if x.native is not y.native:
        raise CheckColumnTypesIntervalEqualError(msg)
    if x.second_precision != y.second_precision:
        raise CheckColumnTypesIntervalEqualError(msg)
    if x.day_precision != y.day_precision:
        raise CheckColumnTypesIntervalEqualError(msg)


class CheckColumnTypesIntervalEqualError(Exception):
    ...


def check_column_types_large_binary_equal(x: Any, y: Any, /) -> None:
    """Check that a pair of large binary column types are equal."""
    if x.length != y.length:
        msg = f"{x=}, {y=}"
        raise CheckColumnTypesLargeBinaryEqualError(msg)


class CheckColumnTypesLargeBinaryEqualError(Exception):
    ...


def check_column_types_numeric_equal(x: Any, y: Any, /) -> None:
    """Check that a pair of numeric column types are equal."""
    msg = f"{x=}, {y=}"
    if x.precision != y.precision:
        raise CheckColumnTypesNumericEqualError(msg)
    if x.scale != y.scale:
        raise CheckColumnTypesNumericEqualError(msg)
    if x.asdecimal != y.asdecimal:
        raise CheckColumnTypesNumericEqualError(msg)
    if x.decimal_return_scale != y.decimal_return_scale:
        raise CheckColumnTypesNumericEqualError(msg)


class CheckColumnTypesNumericEqualError(Exception):
    ...


def check_column_types_string_equal(x: Any, y: Any, /) -> None:
    """Check that a pair of string column types are equal."""
    msg = f"{x=}, {y=}"
    if x.length != y.length:
        raise CheckColumnTypesStringEqualError(msg)
    if x.collation != y.collation:
        raise CheckColumnTypesStringEqualError(msg)


class CheckColumnTypesStringEqualError(Exception):
    ...


def check_column_types_uuid_equal(x: Any, y: Any, /) -> None:
    """Check that a pair of UUID column types are equal."""
    msg = f"{x=}, {y=}"
    if x.as_uuid is not y.as_uuid:
        raise CheckColumnTypesUuidEqualError(msg)
    if x.native_uuid is not y.native_uuid:
        raise CheckColumnTypesUuidEqualError(msg)


class CheckColumnTypesUuidEqualError(Exception):
    ...


def check_engine(
    engine: Engine,
    /,
    *,
    num_tables: IntNonNeg | None = None,
    rel_tol: FloatNonNeg | None = None,
    abs_tol: IntNonNeg | None = None,
) -> None:
    """Check that an engine can connect.

    Optionally query for the number of tables, or the number of columns in
    such a table.
    """
    dialect = get_dialect(engine)
    match dialect:
        case Dialect.mssql | Dialect.mysql | Dialect.postgresql:  # pragma: no cover
            query = "select * from information_schema.tables"
        case Dialect.oracle:  # pragma: no cover
            query = "select * from all_objects"
        case Dialect.sqlite:
            query = "select * from sqlite_master where type='table'"
        case _:  # pragma: no cover  # type: ignore
            assert_never(dialect)

    try:
        with engine.begin() as conn:
            rows = conn.execute(text(query)).all()
    except OperationalError as error:
        redirect_error(error, "unable to open database file", CheckEngineError)
    if num_tables is not None:
        n_rows = len(rows)
        if (rel_tol is None) and (abs_tol is None):
            if n_rows != num_tables:
                msg = f"{len(rows)=}, {num_tables=}"
                raise CheckEngineError(msg)
        else:
            rel_tol_use = 1e-9 if rel_tol is None else rel_tol
            abs_tol_use = 0.0 if abs_tol is None else abs_tol
            if not isclose(
                n_rows, num_tables, rel_tol=rel_tol_use, abs_tol=abs_tol_use
            ):
                msg = f"{len(rows)=}, {num_tables=}, {rel_tol=}, {abs_tol=}"
                raise CheckEngineError(msg)


class CheckEngineError(Exception):
    ...


def check_table_against_reflection(
    table_or_mapped_class: Table | type[Any],
    engine_or_conn: Engine | Connection,
    /,
    *,
    schema: str | None = None,
    snake_table: bool = False,
    snake_columns: bool = False,
    allow_permutations_columns: bool = False,
    primary_key: bool = True,
) -> None:
    """Check that a table equals its reflection."""
    reflected = reflect_table(table_or_mapped_class, engine_or_conn, schema=schema)
    check_tables_equal(
        reflected,
        table_or_mapped_class,
        snake_table=snake_table,
        allow_permutations_columns=allow_permutations_columns,
        snake_columns=snake_columns,
        primary_key=primary_key,
    )


def check_tables_equal(
    x: Any,
    y: Any,
    /,
    *,
    snake_table: bool = False,
    snake_columns: bool = False,
    allow_permutations_columns: bool = False,
    primary_key: bool = True,
) -> None:
    """Check that a pair of tables are equal."""
    x_t, y_t = map(get_table, [x, y])
    check_table_or_column_names_equal(x_t.name, y_t.name, snake=snake_table)
    check_column_collections_equal(
        x_t.columns,
        y_t.columns,
        snake=snake_columns,
        allow_permutations=allow_permutations_columns,
        primary_key=primary_key,
    )


def check_table_or_column_names_equal(
    x: str | quoted_name, y: str | quoted_name, /, *, snake: bool = False
) -> None:
    """Check that a pair of table/columns' names are equal."""
    x, y = (str(i) if isinstance(i, quoted_name) else i for i in [x, y])
    msg = f"{x=}, {y=}"
    if (not snake) and (x != y):
        raise CheckTableOrColumnNamesEqualError(msg)
    if snake and (snake_case(x) != snake_case(y)):
        raise CheckTableOrColumnNamesEqualError(msg)


class CheckTableOrColumnNamesEqualError(Exception):
    ...


def columnwise_max(*columns: Any) -> Any:
    """Compute the columnwise max of a number of columns."""
    return _columnwise_minmax(*columns, op=ge)


def columnwise_min(*columns: Any) -> Any:
    """Compute the columnwise min of a number of columns."""
    return _columnwise_minmax(*columns, op=le)


def _columnwise_minmax(*columns: Any, op: Callable[[Any, Any], Any]) -> Any:
    """Compute the columnwise min of a number of columns."""

    def func(x: Any, y: Any, /) -> Any:
        x_none = x.is_(None)
        y_none = y.is_(None)
        col = case(
            (and_(x_none, y_none), None),
            (and_(~x_none, y_none), x),
            (and_(x_none, ~y_none), y),
            (op(x, y), x),
            else_=y,
        )
        # try auto-label
        names = {
            value for col in [x, y] if (value := getattr(col, "name", None)) is not None
        }
        try:
            (name,) = names
        except ValueError:
            return col
        else:
            return col.label(name)

    return reduce(func, columns)


def create_engine(
    drivername: str,
    /,
    *,
    username: str | None = None,
    password: str | None = None,
    host: str | None = None,
    port: int | None = None,
    database: str | None = None,
    query: Mapping[str, IterableStrs | str] | None = None,
    poolclass: type[Pool] | None = NullPool,
) -> Engine:
    """Create a SQLAlchemy engine."""
    if query is None:
        kwargs = {}
    else:

        def func(x: str | IterableStrs, /) -> list[str] | str:
            return x if isinstance(x, str) else list(x)

        kwargs = {"query": {k: func(v) for k, v in query.items()}}
    url = URL.create(
        drivername,
        username=username,
        password=password,
        host=host,
        port=port,
        database=database,
        **kwargs,
    )
    return _create_engine(url, poolclass=poolclass)


def ensure_engine(engine: Engine | str, /) -> Engine:
    """Ensure the object is an Engine."""
    if isinstance(engine, Engine):
        return engine
    return parse_engine(engine)


def ensure_tables_created(
    engine_or_conn: Engine | Connection, /, *tables_or_mapped_classes: Table | type[Any]
) -> None:
    """Ensure a table/set of tables is/are created."""
    for table_or_mapped_class in tables_or_mapped_classes:
        table = get_table(table_or_mapped_class)
        with yield_connection(engine_or_conn) as conn:
            try:
                table.create(conn)
            except DatabaseError as error:
                with suppress(TableAlreadyExistsError):
                    redirect_to_table_already_exists_error(conn, error)


def ensure_tables_dropped(
    engine_or_conn: Engine | Connection, /, *tables_or_mapped_classes: Table | type[Any]
) -> None:
    """Ensure a table/set of tables is/are dropped."""
    for table_or_mapped_class in tables_or_mapped_classes:
        table = get_table(table_or_mapped_class)
        with yield_connection(engine_or_conn) as conn:
            try:
                table.drop(conn)
            except DatabaseError as error:
                with suppress(NoSuchTableError):
                    redirect_to_no_such_table_error(conn, error)


def get_table_name(table_or_mapped_class: Table | type[Any], /) -> str:
    """Get the table name from a Table or mapped class."""
    return get_table(table_or_mapped_class).name


def parse_engine(engine: str, /) -> Engine:
    """Parse a string into an Engine."""
    try:
        return _create_engine(engine, poolclass=NullPool)
    except ArgumentError:
        raise ParseEngineError from None


class ParseEngineError(Exception):
    ...


def redirect_to_no_such_table_error(
    engine_or_conn: Engine | Connection, error: DatabaseError, /
) -> NoReturn:
    """Redirect to the `NoSuchTableError`."""
    dialect = get_dialect(engine_or_conn)
    match dialect:
        case Dialect.mysql | Dialect.postgresql:  # pragma: no cover
            raise NotImplementedError(dialect)
        case Dialect.mssql:  # pragma: no cover
            pattern = (
                "Cannot drop the table .*, because it does not exist or you do "
                "not have permission"
            )
        case Dialect.oracle:  # pragma: no cover
            pattern = "ORA-00942: table or view does not exist"
        case Dialect.sqlite:
            pattern = "no such table"
        case _:  # pragma: no cover  # type: ignore
            assert_never(dialect)
    return redirect_error(error, pattern, NoSuchTableError)


def redirect_to_table_already_exists_error(
    engine_or_conn: Engine | Connection, error: DatabaseError, /
) -> NoReturn:
    """Redirect to the `TableAlreadyExistsError`."""
    dialect = get_dialect(engine_or_conn)
    match dialect:
        case Dialect.mssql | Dialect.postgresql:  # pragma: no cover
            raise NotImplementedError(dialect)
        case Dialect.mysql:  # pragma: no cover
            pattern = "There is already an object named .* in the database"
        case Dialect.oracle:  # pragma: no cover
            pattern = "ORA-00955: name is already used by an existing object"
        case Dialect.sqlite:
            pattern = "table .* already exists"
        case _:  # pragma: no cover  # type: ignore
            assert_never(dialect)
    return redirect_error(error, pattern, TableAlreadyExistsError)


class TableAlreadyExistsError(Exception):
    ...


def reflect_table(
    table_or_mapped_class: Table | type[Any],
    engine_or_conn: Engine | Connection,
    /,
    *,
    schema: str | None = None,
) -> Table:
    """Reflect a table from a database."""
    name = get_table_name(table_or_mapped_class)
    metadata = MetaData(schema=schema)
    with yield_connection(engine_or_conn) as conn:
        return Table(name, metadata, autoload_with=conn)


def serialize_engine(engine: Engine, /) -> str:
    """Serialize an Engine."""
    return engine.url.render_as_string(hide_password=False)


class TablenameMixin:
    """Mix-in for an auto-generated tablename."""

    @cast(Any, declared_attr)
    def __tablename__(cls) -> str:  # noqa: N805
        return get_class_name(cls, snake=True)


__all__ = [
    "check_column_collections_equal",
    "check_column_types_boolean_equal",
    "check_column_types_datetime_equal",
    "check_column_types_enum_equal",
    "check_column_types_equal",
    "check_column_types_float_equal",
    "check_column_types_interval_equal",
    "check_column_types_large_binary_equal",
    "check_column_types_numeric_equal",
    "check_column_types_uuid_equal",
    "check_columns_equal",
    "check_dataframe_schema_against_table",
    "check_engine",
    "check_selectable_for_duplicate_columns",
    "check_series_against_table_column",
    "check_series_against_table_schema",
    "check_table_against_reflection",
    "check_table_or_column_names_equal",
    "check_tables_equal",
    "CheckColumnCollectionsEqualError",
    "CheckColumnsEqualError",
    "CheckColumnTypesBooleanEqualError",
    "CheckColumnTypesDateTimeEqualError",
    "CheckColumnTypesEnumEqualError",
    "CheckColumnTypesEqualError",
    "CheckColumnTypesFloatEqualError",
    "CheckColumnTypesIntervalEqualError",
    "CheckColumnTypesNumericEqualError",
    "CheckColumnTypesStringEqualError",
    "CheckColumnTypesUuidEqualError",
    "CheckEngineError",
    "CheckSeriesAgainstTableColumnError",
    "CheckSeriesAgainstTableSchemaError",
    "CheckTableOrColumnNamesEqualError",
    "columnwise_max",
    "columnwise_min",
    "create_engine",
    "Dialect",
    "Dialect",
    "ensure_engine",
    "ensure_tables_created",
    "ensure_tables_dropped",
    "get_column_names",
    "get_column_names",
    "get_columns",
    "get_columns",
    "get_dialect",
    "get_dialect",
    "get_table_name",
    "get_table",
    "get_table",
    "GetDialectError",
    "GetTableError",
    "INSERT_ITEMS_CHUNK_SIZE_FRAC",
    "insert_items_collect_iterable",
    "insert_items_collect",
    "insert_items",
    "InsertItemsCollectError",
    "InsertItemsCollectIterableError",
    "is_mapped_class",
    "is_table_or_mapped_class",
    "mapped_class_to_dict",
    "parse_engine",
    "ParseEngineError",
    "redirect_to_no_such_table_error",
    "redirect_to_table_already_exists_error",
    "serialize_engine",
    "TableAlreadyExistsError",
    "TablenameMixin",
    "yield_connection",
]


try:
    from utilities._sqlalchemy.fastparquet import select_to_parquet
except ModuleNotFoundError:  # pragma: no cover
    pass
else:
    __all__ += ["select_to_parquet"]


try:
    from utilities._sqlalchemy.pandas import (
        InsertPandasDataFrameError,
        StreamDataFramesError,
        TableColumnToDTypeError,
        insert_pandas_dataframe,
        select_to_pandas_dataframe,
        stream_dataframes,
        table_column_to_dtype,
    )
except ModuleNotFoundError:  # pragma: no cover
    pass
else:
    __all__ += [
        "insert_pandas_dataframe",
        "InsertPandasDataFrameError",
        "select_to_pandas_dataframe",
        "stream_dataframes",
        "StreamDataFramesError",
        "table_column_to_dtype",
        "TableColumnToDTypeError",
    ]


try:
    from utilities._sqlalchemy.polars import (
        InsertPolarsDataFrameError,
        TableColumnToExprError,
        insert_polars_dataframe,
        select_to_polars_dataframe,
        table_column_to_expr,
    )
except ModuleNotFoundError:  # pragma: no cover
    pass
else:
    __all__ += [
        "insert_polars_dataframe",
        "InsertPolarsDataFrameError",
        "select_to_polars_dataframe",
        "table_column_to_expr",
        "TableColumnToExprError",
    ]

try:
    from utilities._sqlalchemy.timeout_decorator import (
        NextFromSequenceError,
        next_from_sequence,
        redirect_to_next_from_sequence_error,
    )
except ModuleNotFoundError:  # pragma: no cover
    pass
else:
    __all__ += [
        "next_from_sequence",
        "NextFromSequenceError",
        "redirect_to_next_from_sequence_error",
    ]



from __future__ import annotations

from collections.abc import Iterable, Sequence
from functools import reduce
from itertools import chain
from math import isclose

from polars import DataFrame, Expr, PolarsDataType
from polars.exceptions import OutOfBoundsError
from polars.testing import assert_frame_equal
from polars.type_aliases import IntoExpr, JoinStrategy, JoinValidation, SchemaDict

from utilities.types import SequenceStrs


def check_dataframe(
    df: DataFrame,
    /,
    *,
    columns: SequenceStrs | None = None,
    dtypes: list[PolarsDataType] | None = None,
    height: int | tuple[int, float] | None = None,
    min_height: int | None = None,
    max_height: int | None = None,
    schema: SchemaDict | None = None,
    shape: tuple[int, int] | None = None,
    sorted: IntoExpr | Iterable[IntoExpr] | None = None,  # noqa: A002
    unique: IntoExpr | Iterable[IntoExpr] | None = None,
    width: int | None = None,
) -> None:
    """Check the properties of a DataFrame."""
    if (columns is not None) and (df.columns != list(columns)):
        msg = f"{df=}, {columns=}"
        raise CheckDataFrameError(msg)
    if (dtypes is not None) and (df.dtypes != dtypes):
        msg = f"{df=}, {dtypes=}"
        raise CheckDataFrameError(msg)
    if height is not None:
        check_dataframe_height(df, height)
    if (min_height is not None) and (len(df) < min_height):
        msg = f"{df=}, {min_height=}"
        raise CheckDataFrameError(msg)
    if (max_height is not None) and (len(df) > max_height):
        msg = f"{df=}, {max_height=}"
        raise CheckDataFrameError(msg)
    if (schema is not None) and (df.schema != schema):
        set_act, set_exp = map(set, [df.schema, schema])
        extra = set_act - set_exp
        missing = set_exp - set_act
        differ = {
            col: (left, right)
            for col in set_act & set_exp
            if (left := df.schema[col]) != (right := schema[col])
        }
        msg = f"{df=}, {extra=}, {missing=}, {differ=}"
        raise CheckDataFrameError(msg)
    if (shape is not None) and (df.shape != shape):
        msg = f"{df=}"
        raise CheckDataFrameError(msg)
    if sorted is not None:
        df_sorted = df.sort(sorted)
        try:
            assert_frame_equal(df, df_sorted)
        except AssertionError:
            msg = f"{df=}, {sorted=}"
            raise CheckDataFrameError(msg) from None
    if (unique is not None) and df.select(unique).is_duplicated().any():
        msg = f"{df=}, {unique=}"
        raise CheckDataFrameError(msg)
    if (width is not None) and (df.width != width):
        msg = f"{df=}"
        raise CheckDataFrameError(msg)


class CheckDataFrameError(Exception):
    ...


def check_dataframe_height(df: DataFrame, height: int | tuple[int, float], /) -> None:
    """Check the height of a DataFrame."""
    if isinstance(height, int) and (len(df) != height):
        msg = f"{df=}, {height=}"
        raise CheckDataFrameHeightError(msg)
    if isinstance(height, tuple):
        height_int, rel_tol = height
        if not isclose(len(df), height_int, rel_tol=rel_tol):
            msg = f"{df=}, {height=}"
            raise CheckDataFrameHeightError(msg)


class CheckDataFrameHeightError(Exception):
    ...


def join(
    df: DataFrame,
    *dfs: DataFrame,
    on: str | Expr | Sequence[str | Expr],
    how: JoinStrategy = "inner",
    validate: JoinValidation = "m:m",
) -> DataFrame:
    def inner(left: DataFrame, right: DataFrame, /) -> DataFrame:
        return left.join(right, on=on, how=how, validate=validate)

    return reduce(inner, chain([df], dfs))


def set_first_row_as_columns(df: DataFrame, /) -> DataFrame:
    """Set the first row of a DataFrame as its columns."""

    try:
        row = df.row(0)
    except OutOfBoundsError:
        msg = f"{df=}"
        raise SetFirstRowAsColumnsError(msg) from None
    mapping = dict(zip(df.columns, row, strict=True))
    return df[1:].rename(mapping)


class SetFirstRowAsColumnsError(Exception):
    ...


__all__ = [
    "check_dataframe_height",
    "check_dataframe",
    "CheckDataFrameError",
    "CheckDataFrameHeightError",
    "join",
    "set_first_row_as_columns",
    "SetFirstRowAsColumnsError",
]


try:
    from utilities._polars.bs4 import (
        TableTagToDataFrameError,
        table_tag_to_dataframe,
        yield_tables,
    )
except ModuleNotFoundError:  # pragma: no cover
    pass
else:
    __all__ += ["table_tag_to_dataframe", "TableTagToDataFrameError", "yield_tables"]



from __future__ import annotations

from pathlib import Path

from numpy import array, nan
from numpy.testing import assert_equal
from zarr import open_array

from utilities.zarr import ffill_non_nan_slices


class TestFFillNonNanSlices:
    def test_main(self, *, tmp_path: Path) -> None:
        arr = array(
            [[0.1, nan, nan, 0.2], 4 * [nan], [0.3, nan, nan, nan]], dtype=float
        )
        z_arr = open_array(tmp_path.joinpath("array"), shape=arr.shape, dtype=float)
        z_arr[:] = arr
        ffill_non_nan_slices(z_arr)
        expected = array(
            [[0.1, 0.1, 0.1, 0.2], 4 * [nan], [0.3, 0.3, 0.3, nan]], dtype=float
        )
        assert_equal(z_arr[:], expected)



from __future__ import annotations

from collections.abc import Mapping
from typing import TYPE_CHECKING

from hypothesis import given
from hypothesis.strategies import DataObject, data, dictionaries, integers, sampled_from
from xarray import DataArray

from utilities.hypothesis import (
    assume_does_not_raise,
    float_data_arrays,
    int_indexes,
    text_ascii,
)
from utilities.xarray import ewma, exp_moving_sum

if TYPE_CHECKING:  # pragma: no cover
    from utilities.pandas import IndexA


class TestEwma:
    @given(
        data=data(),
        indexes=dictionaries(text_ascii(), int_indexes(), min_size=1, max_size=3),
        halflife=integers(1, 10),
    )
    def test_main(
        self, data: DataObject, indexes: Mapping[str, IndexA], halflife: int
    ) -> None:
        array = data.draw(float_data_arrays(indexes))
        dim = data.draw(sampled_from(list(indexes)))
        with assume_does_not_raise(RuntimeWarning):
            _ = ewma(array, {dim: halflife})


class TestExpMovingSum:
    @given(
        data=data(),
        indexes=dictionaries(text_ascii(), int_indexes(), min_size=1, max_size=3),
        halflife=integers(1, 10),
    )
    def test_main(
        self, data: DataObject, indexes: Mapping[str, IndexA], halflife: int
    ) -> None:
        array = data.draw(float_data_arrays(indexes))
        dim = data.draw(sampled_from(list(indexes)))
        with assume_does_not_raise(RuntimeWarning):
            _ = exp_moving_sum(array, {dim: halflife})


class TestNumbaggInstalled:
    def test_main(self) -> None:
        array = DataArray([], {"dim": []}, ["dim"])
        _ = array.rolling_exp(dim=1.0).sum()



from __future__ import annotations

from re import compile

from utilities.more_itertools import OneError, one


def extract_group(pattern: str, text: str, /) -> str:
    """Extract a group.

    The regex must have 1 capture group, and this must match exactly once.
    """
    compiled = compile(pattern)
    if compiled.groups == 1:
        results = compiled.findall(text)
        try:
            return one(results)
        except OneError:
            msg = f"{pattern=}, {text=}"
            raise ExtractGroupError(msg) from None
    msg = f"{pattern=}, {text=}"
    raise ExtractGroupError(msg)


class ExtractGroupError(Exception):
    ...


def extract_groups(pattern: str, text: str, /) -> list[str]:
    """Extract multiple groups.

    The regex may have any number of capture groups, and they must collectively
    match exactly once.
    """
    compiled = compile(pattern)
    if (n_groups := compiled.groups) >= 1:
        results = compiled.findall(text)
        if len(results) == 1:
            return results if n_groups == 1 else list(results[0])
    msg = f"{pattern=}, {text=}"
    raise ExtractGroupsError(msg)


class ExtractGroupsError(Exception):
    ...


__all__ = ["extract_group", "extract_groups", "ExtractGroupError", "ExtractGroupsError"]



from __future__ import annotations

from typing import Any

from pandas import DatetimeTZDtype, Series
from pytest import mark, param

from utilities.numpy import has_dtype


class TestHasDtype:
    @mark.parametrize(
        ("dtype", "against", "expected"),
        [
            param("Int64", "Int64", True),
            param("Int64", ("Int64",), True),
            param("Int64", int, False),
            param(DatetimeTZDtype(tz="UTC"), DatetimeTZDtype(tz="UTC"), True),
            param(
                DatetimeTZDtype(tz="UTC"), DatetimeTZDtype(tz="Asia/Hong_Kong"), False
            ),
        ],
    )
    def test_main(self, *, dtype: Any, against: Any, expected: bool) -> None:
        result = has_dtype(Series([], dtype=dtype), against)
        assert result is expected



from __future__ import annotations

from hypothesis import given
from hypothesis.strategies import integers, none
from pandas import get_option
from polars import Config

from utilities.jupyter import _DEFAULT_COLS, _DEFAULT_ROWS, show


class TestShow:
    def test_direct(self) -> None:
        expected = {
            "set_fmt_float": "mixed",
            "set_float_precision": None,
            "set_thousands_separator": "",
            "set_decimal_separator": ".",
            "set_trim_decimal_zeros": False,
        }

        def assert_default() -> None:
            assert get_option("display.precision") == 6
            assert get_option("display.min_rows") == 10
            assert get_option("display.max_rows") == 60
            assert get_option("display.max_columns") == 0
            assert Config.state(if_set=True) == expected

        assert_default()
        with show:
            assert get_option("display.precision") == 6
            assert get_option("display.min_rows") == _DEFAULT_ROWS
            assert get_option("display.max_rows") == _DEFAULT_ROWS
            assert get_option("display.max_columns") == _DEFAULT_COLS
            assert Config.state(if_set=True) == expected | {
                "POLARS_FMT_MAX_ROWS": str(_DEFAULT_ROWS),
                "POLARS_FMT_MAX_COLS": str(_DEFAULT_COLS),
            }
        assert_default()

    @given(
        dp=integers(0, 10) | none(),
        rows=integers(0, 100) | none(),
        columns=integers(0, 100) | none(),
    )
    def test_indirect(
        self, *, dp: int | None, rows: int | None, columns: int | None
    ) -> None:
        with show(dp=dp, rows=rows, columns=columns):
            state = Config.state(if_set=True)
            if dp is not None:
                assert get_option("display.precision") == dp
                assert state["set_float_precision"] == dp
            if rows is not None:
                assert get_option("display.min_rows") == rows
                assert get_option("display.max_rows") == rows
                assert state["POLARS_FMT_MAX_ROWS"] == str(rows)
            if columns is not None:
                assert get_option("display.max_columns") == columns
                assert state["POLARS_FMT_MAX_COLS"] == str(columns)



from __future__ import annotations

from polars import DataFrame, Float64, Int64, Utf8
from polars.testing import assert_frame_equal
from pytest import mark, param, raises

from utilities.polars import (
    CheckDataFrameError,
    CheckDataFrameHeightError,
    SetFirstRowAsColumnsError,
    check_dataframe,
    check_dataframe_height,
    join,
    set_first_row_as_columns,
)


class TestCheckDataFrame:
    def test_main(self) -> None:
        df = DataFrame()
        check_dataframe(df)

    def test_columns_pass(self) -> None:
        df = DataFrame()
        check_dataframe(df, columns=[])

    def test_columns_error(self) -> None:
        df = DataFrame()
        with raises(CheckDataFrameError):
            check_dataframe(df, columns=["value"])

    def test_dtypes_pass(self) -> None:
        df = DataFrame()
        check_dataframe(df, dtypes=[])

    def test_dtypes_error(self) -> None:
        df = DataFrame()
        with raises(CheckDataFrameError):
            check_dataframe(df, dtypes=[Float64])

    def test_height(self) -> None:
        df = DataFrame({"value": [0.0]})
        check_dataframe(df, height=1)

    def test_min_height_pass(self) -> None:
        df = DataFrame({"value": [0.0, 1.0]})
        check_dataframe(df, min_height=1)

    def test_min_height_error(self) -> None:
        df = DataFrame()
        with raises(CheckDataFrameError):
            check_dataframe(df, min_height=1)

    def test_max_height_pass(self) -> None:
        df = DataFrame()
        check_dataframe(df, max_height=1)

    def test_max_height_error(self) -> None:
        df = DataFrame({"value": [0.0, 1.0]})
        with raises(CheckDataFrameError):
            check_dataframe(df, max_height=1)

    def test_schema_pass(self) -> None:
        df = DataFrame()
        check_dataframe(df, schema={})

    def test_schema_error(self) -> None:
        df = DataFrame()
        with raises(CheckDataFrameError):
            check_dataframe(df, schema={"value": Float64})

    def test_shape_pass(self) -> None:
        df = DataFrame()
        check_dataframe(df, shape=(0, 0))

    def test_shape_error(self) -> None:
        df = DataFrame()
        with raises(CheckDataFrameError):
            check_dataframe(df, shape=(1, 1))

    def test_sorted_pass(self) -> None:
        df = DataFrame({"value": [0.0, 1.0]})
        check_dataframe(df, sorted="value")

    def test_sorted_error(self) -> None:
        df = DataFrame({"value": [1.0, 0.0]})
        with raises(CheckDataFrameError):
            check_dataframe(df, sorted="value")

    def test_unique_pass(self) -> None:
        df = DataFrame({"value": [0.0, 1.0]})
        check_dataframe(df, unique="value")

    def test_unique_error(self) -> None:
        df = DataFrame({"value": [0.0, 0.0]})
        with raises(CheckDataFrameError):
            check_dataframe(df, unique="value")

    def test_width_pass(self) -> None:
        df = DataFrame()
        check_dataframe(df, width=0)

    def test_width_error(self) -> None:
        df = DataFrame()
        with raises(CheckDataFrameError):
            check_dataframe(df, width=1)


class TestCheckDataFrameHeight:
    @mark.parametrize("height", [param(10), param((11, 0.1))])
    def test_main(self, *, height: int | tuple[int, float]) -> None:
        df = DataFrame({"value": range(10)})
        check_dataframe_height(df, height)

    @mark.parametrize("height", [param(0), param((12, 0.1))])
    def test_error(self, *, height: int | tuple[int, float]) -> None:
        df = DataFrame({"value": range(10)})
        with raises(CheckDataFrameHeightError):
            check_dataframe_height(df, height)


class TestJoin:
    def test_main(self) -> None:
        df1 = DataFrame([{"a": 1, "b": 2}], schema={"a": Int64, "b": Int64})
        df2 = DataFrame([{"a": 1, "c": 3}], schema={"a": Int64, "c": Int64})
        result = join(df1, df2, on="a")
        expected = DataFrame(
            [{"a": 1, "b": 2, "c": 3}], schema={"a": Int64, "b": Int64, "c": Int64}
        )
        assert_frame_equal(result, expected)


class TestSetFirstRowAsColumns:
    def test_empty(self) -> None:
        df = DataFrame()
        with raises(SetFirstRowAsColumnsError):
            _ = set_first_row_as_columns(df)

    def test_one_row(self) -> None:
        df = DataFrame(["value"])
        check_dataframe(df, height=1, schema={"column_0": Utf8})
        result = set_first_row_as_columns(df)
        check_dataframe(result, height=0, schema={"value": Utf8})

    def test_multiple_rows(self) -> None:
        df = DataFrame(["foo", "bar", "baz"])
        check_dataframe(df, height=3, schema={"column_0": Utf8})
        result = set_first_row_as_columns(df)
        check_dataframe(result, height=2, schema={"foo": Utf8})



from __future__ import annotations

from re import escape

from pytest import raises

from utilities.more_itertools import OneEmptyError, OneNonUniqueError, one


class TestOne:
    def test_main(self) -> None:
        assert one([None]) is None

    def test_error_empty(self) -> None:
        with raises(
            OneEmptyError, match=escape("too few items in iterable (expected 1)")
        ):
            _ = one([])

    def test_error_non_unique(self) -> None:
        with raises(
            OneNonUniqueError,
            match="Expected exactly one item in iterable, but got 1, 2, and "
            "perhaps more",
        ):
            _ = one([1, 2, 3])



from __future__ import annotations

from collections.abc import Set as AbstractSet
from math import inf, isfinite, isinf, isnan
from pathlib import Path
from re import search

from hypothesis import given
from hypothesis.errors import InvalidArgument
from hypothesis.strategies import (
    DataObject,
    DrawFn,
    booleans,
    composite,
    data,
    floats,
    integers,
    none,
    sets,
)
from pytest import mark, param

from utilities._hypothesis.common import (
    floats_extra,
    lists_fixed_length,
    temp_dirs,
    temp_paths,
    text_ascii,
)
from utilities.hypothesis import assume_does_not_raise
from utilities.platform import maybe_yield_lower_case
from utilities.tempfile import TemporaryDirectory


class TestFloatsExtra:
    @given(
        data=data(),
        min_value=floats() | none(),
        max_value=floats() | none(),
        allow_nan=booleans(),
        allow_inf=booleans(),
        allow_pos_inf=booleans(),
        allow_neg_inf=booleans(),
        integral=booleans(),
    )
    def test_main(
        self,
        *,
        data: DataObject,
        min_value: float | None,
        max_value: float | None,
        allow_nan: bool,
        allow_inf: bool,
        allow_pos_inf: bool,
        allow_neg_inf: bool,
        integral: bool,
    ) -> None:
        with assume_does_not_raise(InvalidArgument):
            x = data.draw(
                floats_extra(
                    min_value=min_value,
                    max_value=max_value,
                    allow_nan=allow_nan,
                    allow_inf=allow_inf,
                    allow_pos_inf=allow_pos_inf,
                    allow_neg_inf=allow_neg_inf,
                    integral=integral,
                )
            )
        if min_value is not None:
            assert (isfinite(x) and x >= min_value) or not isfinite(x)
        if max_value is not None:
            assert (isfinite(x) and x <= max_value) or not isfinite(x)
        if not allow_nan:
            assert not isnan(x)
        if not allow_inf:
            if not (allow_pos_inf or allow_neg_inf):
                assert not isinf(x)
            if not allow_pos_inf:
                assert x != inf
            if not allow_neg_inf:
                assert x != -inf
        if integral:
            assert (isfinite(x) and x == round(x)) or not isfinite(x)

    @given(data=data(), min_value=floats() | none(), max_value=floats() | none())
    def test_finite_and_integral(
        self, *, data: DataObject, min_value: float | None, max_value: float | None
    ) -> None:  # hard to reach
        with assume_does_not_raise(InvalidArgument):
            x = data.draw(
                floats_extra(
                    min_value=min_value,
                    max_value=max_value,
                    allow_nan=False,
                    allow_inf=False,
                    allow_pos_inf=False,
                    allow_neg_inf=False,
                    integral=True,
                )
            )
        assert isfinite(x)
        if min_value is not None:
            assert x >= min_value
        if max_value is not None:
            assert x <= max_value
        assert x == round(x)


class TestLiftDraw:
    @given(data=data(), x=booleans())
    def test_fixed(self, *, data: DataObject, x: bool) -> None:
        @composite
        def func(_draw: DrawFn, /) -> bool:
            _ = _draw(booleans())
            return x

        result = data.draw(func())
        assert result is x

    @given(data=data())
    def test_strategy(self, *, data: DataObject) -> None:
        @composite
        def func(_draw: DrawFn, /) -> bool:
            return _draw(booleans())

        result = data.draw(func())
        assert isinstance(result, bool)


class TestListsFixedLength:
    @given(data=data(), size=integers(1, 10))
    @mark.parametrize(
        "unique", [param(True, id="unique"), param(False, id="no unique")]
    )
    @mark.parametrize(
        "sorted_", [param(True, id="sorted"), param(False, id="no sorted")]
    )
    def test_main(
        self, *, data: DataObject, size: int, unique: bool, sorted_: bool
    ) -> None:
        result = data.draw(
            lists_fixed_length(integers(), size, unique=unique, sorted=sorted_)
        )
        assert isinstance(result, list)
        assert len(result) == size
        if unique:
            assert len(set(result)) == len(result)
        if sorted_:
            assert sorted(result) == result


class TestTempDirs:
    @given(temp_dir=temp_dirs())
    def test_main(self, *, temp_dir: TemporaryDirectory) -> None:
        path = temp_dir.path
        assert path.is_dir()
        assert len(set(path.iterdir())) == 0

    @given(temp_dir=temp_dirs(), contents=sets(text_ascii(min_size=1), max_size=10))
    def test_writing_files(
        self, *, temp_dir: TemporaryDirectory, contents: AbstractSet[str]
    ) -> None:
        path = temp_dir.path
        assert len(set(path.iterdir())) == 0
        as_set = set(maybe_yield_lower_case(contents))
        for content in as_set:
            path.joinpath(content).touch()
        assert len(set(path.iterdir())) == len(as_set)


class TestTempPaths:
    @given(path=temp_paths())
    def test_main(self, *, path: Path) -> None:
        assert path.is_dir()
        assert len(set(path.iterdir())) == 0

    @given(path=temp_paths(), contents=sets(text_ascii(min_size=1), max_size=10))
    def test_writing_files(self, *, path: Path, contents: AbstractSet[str]) -> None:
        assert len(set(path.iterdir())) == 0
        as_set = set(maybe_yield_lower_case(contents))
        for content in as_set:
            path.joinpath(content).touch()
        assert len(set(path.iterdir())) == len(as_set)


class TestTextAscii:
    @given(
        data=data(),
        min_size=integers(0, 100),
        max_size=integers(0, 100) | none(),
        disallow_na=booleans(),
    )
    def test_main(
        self,
        *,
        data: DataObject,
        min_size: int,
        max_size: int | None,
        disallow_na: bool,
    ) -> None:
        with assume_does_not_raise(InvalidArgument, AssertionError):
            text = data.draw(
                text_ascii(
                    min_size=min_size, max_size=max_size, disallow_na=disallow_na
                )
            )
        assert search("^[A-Za-z]*$", text)
        assert len(text) >= min_size
        if max_size is not None:
            assert len(text) <= max_size
        if disallow_na:
            assert text != "NA"



from __future__ import annotations

from collections.abc import Iterator
from contextlib import contextmanager
from pathlib import Path
from shutil import move, rmtree

from atomicwrites import move_atomic, replace_atomic
from pathvalidate import validate_filepath

from utilities.errors import DirectoryExistsError
from utilities.pathlib import PathLike
from utilities.tempfile import TemporaryDirectory


@contextmanager
def writer(path: PathLike, /, *, overwrite: bool = False) -> Iterator[Path]:
    """Yield a path for atomically writing files to disk."""
    path = Path(path)
    validate_filepath(path, platform="auto")
    parent = path.parent
    parent.mkdir(parents=True, exist_ok=True)
    name = path.name
    with TemporaryDirectory(suffix=".tmp", prefix=name, dir=parent) as temp_dir:
        try:
            yield (temp_path := temp_dir.joinpath(name))
        except KeyboardInterrupt:
            rmtree(temp_dir)
        else:
            if temp_path.is_file():
                src, dest = map(str, [temp_path, path])
                if overwrite:
                    return replace_atomic(src, dest)
                return move_atomic(src, dest)
            elif temp_path.is_dir():
                if (not path.exists()) or overwrite:
                    return move(temp_path, path)
                msg = f"{temp_dir=}, {path=}"
                raise DirectoryExistsError(msg)
            else:
                msg = f"{temp_path=}"
                raise WriterError(msg)


class WriterError(Exception):
    ...


__all__ = ["writer", "WriterError"]



from __future__ import annotations

from typing import cast

from bottleneck import push
from numpy import errstate, flip, isfinite, nan, where

from utilities._numpy.common import NDArrayF, NDArrayI, shift


def ffill(array: NDArrayF, /, *, limit: int | None = None, axis: int = -1) -> NDArrayF:
    """Forward fill the elements in an array."""
    return push(array, n=limit, axis=axis)


def pct_change(
    array: NDArrayF | NDArrayI,
    /,
    *,
    limit: int | None = None,
    n: int = 1,
    axis: int = -1,
) -> NDArrayF:
    """Compute the percentage change in an array."""
    if n == 0:
        msg = f"{n=}"
        raise PctChangeError(msg)
    if n > 0:
        filled = ffill(array.astype(float), limit=limit, axis=axis)
        shifted = shift(filled, n=n, axis=axis)
        with errstate(all="ignore"):
            ratio = (filled / shifted) if n >= 0 else (shifted / filled)
        return where(isfinite(array), ratio - 1.0, nan)
    flipped = cast(NDArrayF | NDArrayI, flip(array, axis=axis))
    result = pct_change(flipped, limit=limit, n=-n, axis=axis)
    return flip(result, axis=axis)


class PctChangeError(Exception):
    ...


__all__ = ["ffill", "pct_change", "PctChangeError"]



import datetime as dt
from collections.abc import Callable
from dataclasses import dataclass
from operator import eq
from pathlib import Path
from typing import Any, TypeVar

from hypothesis import given
from hypothesis.strategies import (
    DataObject,
    SearchStrategy,
    data,
    dates,
    datetimes,
    just,
    timedeltas,
    times,
    tuples,
)
from pytest import mark, param, raises
from typed_settings.exceptions import InvalidSettingsError

from utilities.datetime import (
    UTC,
    serialize_date,
    serialize_datetime,
    serialize_time,
    serialize_timedelta,
)
from utilities.hypothesis import temp_paths, text_ascii
from utilities.typed_settings import load_settings

app_names = text_ascii(min_size=1).map(str.lower)


_T = TypeVar("_T")


class TestLoadSettings:
    @given(data=data(), root=temp_paths(), appname=app_names)
    @mark.parametrize(
        ("test_cls", "strategy", "serialize"),
        [
            param(dt.date, dates(), serialize_date),
            param(dt.datetime, datetimes(timezones=just(UTC)), serialize_datetime),
            param(dt.time, times(), serialize_time),
            param(dt.timedelta, timedeltas(), serialize_timedelta),
        ],
    )
    def test_main(
        self,
        *,
        data: DataObject,
        root: Path,
        appname: str,
        test_cls: type[_T],
        strategy: SearchStrategy[_T],
        serialize: Callable[[_T], str],
    ) -> None:
        default, value = data.draw(tuples(strategy, strategy))
        self.run_test(test_cls, default, root, appname, serialize, value, eq)

    @staticmethod
    def run_test(
        test_cls: type[_T],
        default: _T,
        root: Path,
        appname: str,
        serialize: Callable[[_T], str],
        value: _T,
        equal: Callable[[_T, _T], bool],
        /,
    ) -> None:
        @dataclass(frozen=True)
        class Settings:
            value: test_cls = default

        settings_default = load_settings(Settings)
        assert settings_default.value == default
        _ = hash(settings_default)
        file = root.joinpath("file.toml")
        with file.open(mode="w") as fh:
            _ = fh.write(f'[{appname}]\nvalue = "{serialize(value)}"')
        settings_loaded = load_settings(Settings, appname=appname, config_files=[file])
        assert equal(settings_loaded.value, value)

    @given(appname=app_names)
    @mark.parametrize("cls", [param(dt.date), param(dt.time), param(dt.timedelta)])
    def test_errors(self, *, appname: str, cls: Any) -> None:
        @dataclass(frozen=True)
        class Settings:
            value: cls = None

        with raises(InvalidSettingsError):
            _ = load_settings(Settings, appname=appname)



from __future__ import annotations

from operator import sub

from hypothesis import given
from hypothesis.strategies import integers

from utilities.functools import partial


class TestPartial:
    @given(x=integers(), y=integers())
    def test_main(self, *, x: int, y: int) -> None:
        func = partial(sub, ..., y)
        assert func(x) == x - y



from __future__ import annotations

from typing import NoReturn

from pytest import mark, param, raises

from utilities.errors import RedirectErrorError, redirect_error, retry
from utilities.more_itertools import one


class TestRedirectError:
    class _CustomError(Exception):
        ...

    def test_generic_redirected_to_custom(self) -> None:
        with raises(self._CustomError):
            self._raises_custom("generic error")

    def test_generic_not_redirected_to_custom(self) -> None:
        with raises(ValueError, match="generic error"):
            self._raises_custom("something else")

    def _raises_custom(self, pattern: str, /) -> NoReturn:
        try:
            msg = "generic error"
            raise ValueError(msg)  # noqa: TRY301
        except ValueError as error:
            redirect_error(error, pattern, self._CustomError)

    def test_generic_with_no_unique_arg(self) -> None:
        with raises(RedirectErrorError):
            try:
                raise ValueError(0, 1)  # noqa: TRY301
            except ValueError as error:
                redirect_error(error, "error", RuntimeError)


class TestRetry:
    @mark.parametrize("use_predicate", [param(None), param(True), param(False)])
    def test_main(self, *, use_predicate: bool | None) -> None:
        class TooLargeError(Exception):
            ...

        def increment() -> int:
            nonlocal n
            n += 1
            if n >= 3:
                raise TooLargeError(n)
            return n

        n = 0
        assert increment() == 1
        assert increment() == 2
        with raises(TooLargeError):
            _ = increment()

        def reset(_error: TooLargeError, /) -> None:
            nonlocal n
            n = 0

        if use_predicate is None:
            retry_inc = retry(increment, TooLargeError, reset)
        else:

            def predicate(error: TooLargeError, /) -> bool:
                if use_predicate:
                    return one(error.args) >= 3
                return one(error.args) >= 4

            retry_inc = retry(increment, TooLargeError, reset, predicate=predicate)

        n = 0
        assert retry_inc() == 1
        assert retry_inc() == 2
        if (use_predicate is None) or (use_predicate is True):
            assert retry_inc() == 1
        else:
            with raises(TooLargeError):
                _ = retry_inc()



from __future__ import annotations

from collections.abc import Iterable, Mapping
from dataclasses import dataclass
from io import StringIO, TextIOWrapper
from typing import Any, cast

from tqdm import tqdm as _tqdm

from utilities.pytest import is_pytest

_LockArgs = tuple[bool | None, float | None] | tuple[bool | None] | None


@dataclass(frozen=True)
class _Defaults:
    desc: str | None = None
    total: float | None = None
    leave: bool | None = True
    file: TextIOWrapper | StringIO | None = None
    ncols: int | None = None
    mininterval: float | None = 0.1
    maxinterval: float | None = 10.0
    miniters: int | float | None = None
    ascii: bool | str | None = None  # noqa: A003
    unit: str | None = "i"
    unit_scale: bool | int | str | None = False
    dynamic_ncols: bool | None = True
    smoothing: float | None = 0.3
    bar_format: str | None = (
        "{desc}: {percentage:3.0f}% | "
        "{elapsed} +{remaining} ={eta:%H:%M:%S} | "
        "{n}/{total} | {rate_fmt}"
    )
    initial: int | float | None = 0
    position: int | None = None
    postfix: Mapping[str, Any] | None = None
    unit_divisor: float | None = 1000.0
    write_bytes: bool | None = None
    lock_args: _LockArgs = None
    nrows: int | None = None
    colour: str | None = None
    delay: float | None = 0.0
    gui: bool | None = False


_DEFAULTS = _Defaults()


class tqdm(_tqdm):  # noqa: N801
    """Sub-class of `tqdm` which is disabled during pytest."""

    def __init__(
        self,
        iterable: Iterable[Any] | None = None,
        desc: str | None = _DEFAULTS.desc,
        total: float | None = _DEFAULTS.total,
        leave: bool | None = _DEFAULTS.leave,
        file: TextIOWrapper | StringIO | None = _DEFAULTS.file,
        ncols: int | None = _DEFAULTS.ncols,
        mininterval: float | None = _DEFAULTS.mininterval,
        maxinterval: float | None = _DEFAULTS.maxinterval,
        miniters: float | None = _DEFAULTS.miniters,
        ascii: bool | str | None = None,  # noqa: A002
        unit: str | None = _DEFAULTS.unit,
        unit_scale: bool | int | str | None = _DEFAULTS.unit_scale,
        dynamic_ncols: bool | None = _DEFAULTS.dynamic_ncols,
        smoothing: float | None = _DEFAULTS.smoothing,
        bar_format: str | None = _DEFAULTS.bar_format,
        initial: float | None = 0,
        position: int | None = _DEFAULTS.position,
        postfix: Mapping[str, Any] | None = _DEFAULTS.postfix,
        unit_divisor: float | None = _DEFAULTS.unit_divisor,
        write_bytes: bool | None = _DEFAULTS.write_bytes,
        lock_args: _LockArgs = _DEFAULTS.lock_args,
        nrows: int | None = _DEFAULTS.nrows,
        colour: str | None = _DEFAULTS.colour,
        delay: float | None = _DEFAULTS.delay,
        gui: bool | None = _DEFAULTS.gui,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            iterable=cast(Any, iterable),
            desc=desc,
            total=_get_total(total, iterable),
            leave=leave,
            file=file,
            ncols=ncols,
            mininterval=cast(Any, mininterval),
            maxinterval=cast(Any, maxinterval),
            miniters=miniters,
            ascii=ascii,
            disable=is_pytest(),
            unit=cast(Any, unit),
            unit_scale=cast(Any, unit_scale),
            dynamic_ncols=cast(Any, dynamic_ncols),
            smoothing=cast(Any, smoothing),
            bar_format=bar_format,
            initial=cast(Any, initial),
            position=position,
            postfix=postfix,
            unit_divisor=cast(Any, unit_divisor),
            write_bytes=write_bytes,
            lock_args=lock_args,
            nrows=nrows,
            colour=colour,
            delay=delay,
            gui=cast(Any, gui),
            **kwargs,
        )


def _get_total(total: float | None, iterable: Any, /) -> float | None:
    if total is not None:
        return total
    try:
        return len(iterable)
    except TypeError:
        return None


__all__ = ["tqdm"]



from __future__ import annotations

from collections.abc import Iterator, Mapping
from functools import partial
from itertools import chain, repeat, starmap
from pathlib import Path
from re import MULTILINE, escape, search
from subprocess import PIPE, CalledProcessError, check_output
from typing import Any

from utilities.more_itertools import OneError, one
from utilities.os import temp_environ
from utilities.pathlib import PathLike
from utilities.types import IterableStrs


def get_shell_output(
    cmd: str,
    /,
    *,
    cwd: PathLike = Path.cwd(),
    activate: PathLike | None = None,
    env: Mapping[str, str | None] | None = None,
) -> str:
    """Get the output of a shell call.

    Optionally, activate a virtual environment if necessary.
    """
    cwd = Path(cwd)
    if activate is not None:
        try:
            activate = one(cwd.rglob("activate"))
        except OneError:
            msg = f"{cwd=}"
            raise GetShellOutputError(msg) from None
        cmd = f"source {activate}; {cmd}"  # pragma: os-ne-windows

    with temp_environ(env):
        return check_output(
            cmd,
            stderr=PIPE,
            shell=True,  # noqa: S602
            cwd=cwd,
            text=True,
        )


class GetShellOutputError(Exception):
    ...


def run_accept_address_in_use(args: IterableStrs, /, *, exist_ok: bool) -> None:
    """Run a command, accepting the 'address already in use' error."""
    try:  # pragma: no cover
        _ = check_output(list(args), stderr=PIPE, text=True)  # noqa: S603
    except CalledProcessError as error:  # pragma: no cover
        pattern = _address_already_in_use_pattern()
        try:
            from loguru import logger
        except ModuleNotFoundError:
            info = exception = print
        else:
            info = logger.info
            exception = logger.exception
        if exist_ok and search(pattern, error.stderr, flags=MULTILINE):
            info("Address already in use")
        else:
            exception("Address already in use")
            raise


def _address_already_in_use_pattern() -> str:
    """Get the 'address_already_in_use' pattern."""
    text = "OSError: [Errno 98] Address already in use"
    escaped = escape(text)
    return f"^{escaped}$"


def tabulate_called_process_error(error: CalledProcessError, /) -> str:
    """Tabulate the components of a CalledProcessError."""
    mapping = {  # pragma: os-ne-windows
        "cmd": error.cmd,
        "returncode": error.returncode,
        "stdout": error.stdout,
        "stderr": error.stderr,
    }
    max_key_len = max(map(len, mapping))  # pragma: os-ne-windows
    tabulate = partial(_tabulate, buffer=max_key_len + 1)  # pragma: os-ne-windows
    return "\n".join(starmap(tabulate, mapping.items()))  # pragma: os-ne-windows


def _tabulate(key: str, value: Any, /, *, buffer: int) -> str:
    template = f"{{:{buffer}}}{{}}"  # pragma: os-ne-windows

    def yield_lines() -> Iterator[str]:  # pragma: os-ne-windows
        keys = chain([key], repeat(buffer * " "))
        value_lines = str(value).splitlines()
        for k, v in zip(keys, value_lines, strict=False):
            yield template.format(k, v)

    return "\n".join(yield_lines())  # pragma: os-ne-windows


__all__ = [
    "get_shell_output",
    "GetShellOutputError",
    "run_accept_address_in_use",
    "tabulate_called_process_error",
]



from __future__ import annotations

import datetime as dt
from types import NoneType
from typing import Any

from beartype.door import die_if_unbearable
from beartype.roar import BeartypeAbbyHintViolation
from pytest import mark, param, raises

from utilities.types import (
    Duration,
    EnsureHashableError,
    IterableStrs,
    Number,
    SequenceStrs,
    ensure_class,
    ensure_hashable,
    is_hashable,
    is_sized_not_str,
    issubclass_except_bool_int,
)


class TestDuration:
    @mark.parametrize("x", [param(0), param(0.0), param(dt.timedelta(0))])
    def test_success(self, *, x: Duration) -> None:
        die_if_unbearable(x, Duration)

    def test_error(self) -> None:
        with raises(BeartypeAbbyHintViolation):
            die_if_unbearable("0", Duration)


class TestEnsureClass:
    @mark.parametrize(
        ("obj", "expected"), [param(None, NoneType), param(NoneType, NoneType)]
    )
    def test_main(self, *, obj: Any, expected: type[Any]) -> None:
        assert ensure_class(obj) is expected


class TestEnsureHashable:
    @mark.parametrize("obj", [param(0), param((1, 2, 3))])
    def test_main(self, *, obj: Any) -> None:
        assert ensure_hashable(obj) == obj

    def test_error(self) -> None:
        with raises(EnsureHashableError):
            _ = ensure_hashable([1, 2, 3])


class TestIsHashable:
    @mark.parametrize(
        ("obj", "expected"),
        [param(0, True), param((1, 2, 3), True), param([1, 2, 3], False)],
    )
    def test_main(self, *, obj: Any, expected: bool) -> None:
        assert is_hashable(obj) is expected


class TestIsSizedNotStr:
    @mark.parametrize(
        ("obj", "expected"),
        [param(None, False), param([], True), param((), True), param("", False)],
    )
    def test_main(self, *, obj: Any, expected: bool) -> None:
        assert is_sized_not_str(obj) is expected


class TestIsSubclassExceptBoolInt:
    @mark.parametrize(
        ("x", "y", "expected"),
        [param(bool, bool, True), param(bool, int, False), param(int, int, True)],
    )
    def test_main(self, *, x: type[Any], y: type[Any], expected: bool) -> None:
        assert issubclass_except_bool_int(x, y) is expected

    def test_subclass_of_int(self) -> None:
        class MyInt(int):
            ...

        assert not issubclass_except_bool_int(bool, MyInt)


class TestIterableStrs:
    @mark.parametrize(
        "x",
        [
            param(["a", "b", "c"]),
            param(("a", "b", "c")),
            param({"a", "b", "c"}),
            param({"a": 1, "b": 2, "c": 3}),
        ],
    )
    def test_pass(self, *, x: IterableStrs) -> None:
        die_if_unbearable(x, IterableStrs)

    def test_fail(self) -> None:
        with raises(BeartypeAbbyHintViolation):
            die_if_unbearable("abc", IterableStrs)


class TestNumber:
    @mark.parametrize("x", [param(0), param(0.0)])
    def test_ok(self, *, x: Number) -> None:
        die_if_unbearable(x, Number)

    def test_error(self) -> None:
        with raises(BeartypeAbbyHintViolation):
            die_if_unbearable("0", Number)


class TestSequenceStrs:
    @mark.parametrize("x", [param(["a", "b", "c"]), param(("a", "b", "c"))])
    def test_pass(self, *, x: SequenceStrs) -> None:
        die_if_unbearable(x, SequenceStrs)

    @mark.parametrize(
        "x", [param({"a", "b", "c"}), param({"a": 1, "b": 2, "c": 3}), param("abc")]
    )
    def test_fail(self, *, x: IterableStrs | str) -> None:
        with raises(BeartypeAbbyHintViolation):
            die_if_unbearable(x, SequenceStrs)



from __future__ import annotations

from collections.abc import Iterable
from typing import Any, TypeVar, cast

from typed_settings import load_settings as _load_settings
from typed_settings.types import AUTO, _Auto

from utilities._typed_settings.common import (
    CONFIG_FILES,
    ExtendedTSConverter,
    GetLoadersError,
    get_loaders,
)
from utilities.pathlib import PathLike

_T = TypeVar("_T")


def load_settings(
    cls: type[_T],
    /,
    *,
    appname: str = "appname",
    config_files: Iterable[PathLike] = CONFIG_FILES,
    config_file_section: str | _Auto = AUTO,
    config_files_var: None | str | _Auto = AUTO,
    env_prefix: None | str | _Auto = AUTO,
) -> _T:
    """Load a settings object with the extended converter."""
    loaders = get_loaders(
        appname=appname,
        config_files=config_files,
        config_file_section=config_file_section,
        config_files_var=config_files_var,
        env_prefix=env_prefix,
    )
    converter = ExtendedTSConverter()
    return _load_settings(cast(Any, cls), loaders, converter=converter)


__all__ = [
    "CONFIG_FILES",
    "ExtendedTSConverter",
    "get_loaders",
    "GetLoadersError",
    "load_settings",
]


try:
    from utilities._typed_settings.click import click_field, click_options
except ModuleNotFoundError:  # pragma: no cover
    pass
else:
    __all__ += ["click_field", "click_options"]



from __future__ import annotations

import datetime as dt
from collections.abc import Hashable, Mapping, Sequence
from functools import partial, reduce
from itertools import permutations
from math import isclose
from typing import TYPE_CHECKING, Any, Literal, NoReturn, TypeAlias, TypeVar, cast

from numpy import where
from pandas import (
    NA,
    BooleanDtype,
    CategoricalDtype,
    DataFrame,
    DatetimeTZDtype,
    Index,
    Int64Dtype,
    NaT,
    RangeIndex,
    Series,
    StringDtype,
    Timestamp,
)
from pandas.testing import assert_frame_equal, assert_index_equal

from utilities.datetime import UTC
from utilities.errors import redirect_error
from utilities.numpy import NDArray1, dt64ns, has_dtype
from utilities.zoneinfo import HONG_KONG

if TYPE_CHECKING:  # pragma: no cover
    IndexA: TypeAlias = Index[Any]
    IndexB: TypeAlias = Index[bool]
    IndexBn: TypeAlias = Index[BooleanDtype]
    IndexC: TypeAlias = Index[CategoricalDtype]
    IndexD: TypeAlias = Index[dt.datetime]
    IndexDhk: TypeAlias = Index[DatetimeTZDtype]
    IndexDutc: TypeAlias = Index[DatetimeTZDtype]
    IndexF: TypeAlias = Index[float]
    IndexI: TypeAlias = Index[int]
    IndexI64: TypeAlias = Index[Int64Dtype]
    IndexS: TypeAlias = Index[StringDtype]

    SeriesA: TypeAlias = Series[Any]
    SeriesB: TypeAlias = Series[bool]
    SeriesBn: TypeAlias = Series[BooleanDtype]
    SeriesC: TypeAlias = Series[CategoricalDtype]
    SeriesD: TypeAlias = Series[dt.datetime]
    SeriesDhk: TypeAlias = Series[DatetimeTZDtype]
    SeriesDutc: TypeAlias = Series[DatetimeTZDtype]
    SeriesF: TypeAlias = Series[float]
    SeriesI: TypeAlias = Series[int]
    SeriesI64: TypeAlias = Series[Int64Dtype]
    SeriesS: TypeAlias = Series[StringDtype]


Int64 = "Int64"
boolean = "boolean"
category = "category"
string = "string"
datetime64nsutc = DatetimeTZDtype(tz=UTC)
datetime64nshk = DatetimeTZDtype(tz=HONG_KONG)


_Index = TypeVar("_Index", bound=Index)


def astype(df: DataFrame, dtype: Any, /) -> DataFrame:
    """Wrapper around `.astype`."""
    return cast(Any, df).astype(dtype)


def check_dataframe(
    df: DataFrame,
    /,
    *,
    columns: Sequence[Hashable] | None = None,
    dtypes: Mapping[Hashable, Any] | None = None,
    length: int | tuple[int, float] | None = None,
    min_length: int | None = None,
    max_length: int | None = None,
    sorted: Hashable | Sequence[Hashable] | None = None,  # noqa: A002
    unique: Hashable | Sequence[Hashable] | None = None,
) -> None:
    """Check the properties of a DataFrame."""
    check_range_index(df.index)
    if df.columns.name is not None:
        msg = f"{df=}"
        raise CheckDataFrameError(msg)
    if df.columns.duplicated().any():
        msg = f"{df=}"
        raise CheckDataFrameError(msg)
    if (columns is not None) and (list(df.columns) != columns):
        msg = f"{df=}, {columns=}"
        raise CheckDataFrameError(msg)
    if (dtypes is not None) and (dict(df.dtypes) != dict(dtypes)):
        msg = f"{df=}, {dtypes=}"
        raise CheckDataFrameError(msg)
    if length is not None:
        check_dataframe_length(df, length)
    if (min_length is not None) and (len(df) < min_length):
        msg = f"{df=}, {min_length=}"
        raise CheckDataFrameError(msg)
    if (max_length is not None) and (len(df) > max_length):
        msg = f"{df=}, {max_length=}"
        raise CheckDataFrameError(msg)
    if sorted is not None:
        df_sorted: DataFrame = df.sort_values(by=sorted).reset_index(drop=True)  # type: ignore
        try:
            assert_frame_equal(df, df_sorted)
        except AssertionError:
            msg = f"{df=}, {sorted=}"
            raise CheckDataFrameError(msg) from None
    if (unique is not None) and df.duplicated(subset=unique).any():
        msg = f"{df=}, {unique=}"
        raise CheckDataFrameError(msg)


class CheckDataFrameError(Exception):
    ...


def check_dataframe_length(df: DataFrame, length: int | tuple[int, float], /) -> None:
    """Check the length of a DataFrame."""
    if isinstance(length, int) and (len(df) != length):
        msg = f"{df=}, {length=}"
        raise CheckDataFrameLengthError(msg)
    if isinstance(length, tuple):
        length_int, rel_tol = length
        if not isclose(len(df), length_int, rel_tol=rel_tol):
            msg = f"{df=}, {length=}"
            raise CheckDataFrameLengthError(msg)


class CheckDataFrameLengthError(Exception):
    ...


def check_range_index(obj: IndexA | SeriesA | DataFrame, /) -> None:
    """Check if a RangeIndex is the default one."""
    msg = f"{obj=}"
    if isinstance(obj, Index):
        if not isinstance(obj, RangeIndex):
            raise CheckRangeIndexError(msg)
        if obj.start != 0:  # type: ignore
            raise CheckRangeIndexError(msg)
        if obj.step != 1:  # type: ignore
            raise CheckRangeIndexError(msg)
        if obj.name is not None:
            raise CheckRangeIndexError(msg)
    else:
        check_range_index(obj.index)


class CheckRangeIndexError(Exception):
    ...


def redirect_to_empty_pandas_concat_error(error: ValueError, /) -> NoReturn:
    """Redirect to the `EmptyPandasConcatError`."""
    redirect_error(error, "No objects to concatenate", EmptyPandasConcatError)


class EmptyPandasConcatError(Exception):
    ...


def rename_index(index: _Index, name: Hashable, /) -> _Index:
    """Wrapper around `.rename`."""
    return cast(_Index, index.rename(name))


def series_max(*series: SeriesA) -> SeriesA:
    """Compute the maximum of a set of Series."""
    return reduce(partial(series_minmax, kind="lower"), series)


def series_min(*series: SeriesA) -> SeriesA:
    """Compute the minimum of a set of Series."""
    return reduce(partial(series_minmax, kind="upper"), series)


def series_minmax(
    x: SeriesA, y: SeriesA, /, *, kind: Literal["lower", "upper"]
) -> SeriesA:
    """Compute the minimum/maximum of a pair of Series."""
    assert_index_equal(x.index, y.index)
    if not (has_dtype(x, y.dtype) and has_dtype(y, x.dtype)):
        msg = f"{x.dtype=}, {y.dtype=}"
        raise SeriesMinMaxError(msg)
    out = x.copy()
    for first, second in permutations([x, y]):
        i = first.notna() & second.isna()
        out.loc[i] = first.loc[i]
    i = x.notna() & y.notna()
    out.loc[i] = x.loc[i].clip(**{kind: cast(Any, y.loc[i])})
    out.loc[x.isna() & y.isna()] = NA
    return out


class SeriesMinMaxError(Exception):
    ...


def sort_index(index: _Index, /) -> _Index:
    return cast(_Index, index.sort_values())


def timestamp_to_date(timestamp: Any, /, *, warn: bool = True) -> dt.date:
    """Convert a timestamp to a date."""
    return timestamp_to_datetime(timestamp, warn=warn).date()


def timestamp_to_datetime(timestamp: Any, /, *, warn: bool = True) -> dt.datetime:
    """Convert a timestamp to a datetime."""
    if timestamp is NaT:
        msg = f"{timestamp=}"
        raise TimestampToDateTimeError(msg)
    datetime = cast(dt.datetime, timestamp.to_pydatetime(warn=warn))
    if datetime.tzinfo is None:
        return datetime.replace(tzinfo=UTC)
    return datetime


class TimestampToDateTimeError(Exception):
    ...


def _timestamp_minmax_to_date(timestamp: Timestamp, method_name: str, /) -> dt.date:
    """Get the maximum Timestamp as a date."""
    method = getattr(timestamp, method_name)
    rounded = cast(Timestamp, method("D"))
    return timestamp_to_date(rounded)


TIMESTAMP_MIN_AS_DATE = _timestamp_minmax_to_date(Timestamp.min, "ceil")
TIMESTAMP_MAX_AS_DATE = _timestamp_minmax_to_date(Timestamp.max, "floor")


def _timestamp_minmax_to_datetime(
    timestamp: Timestamp, method_name: str, /
) -> dt.datetime:
    """Get the maximum Timestamp as a datetime."""
    method = getattr(timestamp, method_name)
    rounded = cast(Timestamp, method("us"))
    return timestamp_to_datetime(rounded)


TIMESTAMP_MIN_AS_DATETIME = _timestamp_minmax_to_datetime(Timestamp.min, "ceil")
TIMESTAMP_MAX_AS_DATETIME = _timestamp_minmax_to_datetime(Timestamp.max, "floor")


def to_numpy(series: SeriesA, /) -> NDArray1:
    """Convert a series into a 1-dimensional `ndarray`."""
    if has_dtype(series, (bool, dt64ns, int, float)):
        return series.to_numpy()
    if has_dtype(series, (boolean, Int64, string)):
        return where(
            series.notna(), series.to_numpy(dtype=object), cast(Any, None)
        ).astype(object)
    msg = f"{series=}"  # pragma: no cover
    raise ToNumpyError(msg)  # pragma: no cover


class ToNumpyError(Exception):
    ...


__all__ = [
    "astype",
    "boolean",
    "category",
    "check_dataframe_length",
    "check_dataframe",
    "check_range_index",
    "CheckDataFrameError",
    "CheckDataFrameLengthError",
    "CheckRangeIndexError",
    "datetime64nshk",
    "datetime64nsutc",
    "EmptyPandasConcatError",
    "Int64",
    "redirect_to_empty_pandas_concat_error",
    "rename_index",
    "series_max",
    "series_min",
    "series_minmax",
    "SeriesMinMaxError",
    "sort_index",
    "string",
    "TIMESTAMP_MAX_AS_DATE",
    "TIMESTAMP_MAX_AS_DATETIME",
    "TIMESTAMP_MIN_AS_DATE",
    "TIMESTAMP_MIN_AS_DATETIME",
    "timestamp_to_date",
    "timestamp_to_datetime",
    "TimestampToDateTimeError",
    "to_numpy",
]



from __future__ import annotations

from pytest import raises

from utilities._typed_settings.common import get_loaders
from utilities.typed_settings import GetLoadersError


class TestGetLoaders:
    def test_success(self) -> None:
        _ = get_loaders()

    def test_error(self) -> None:
        with raises(GetLoadersError):
            _ = get_loaders(appname="app_name")



from __future__ import annotations

from functools import partial as _partial
from typing import Any, TypeVar

from typing_extensions import override

_T = TypeVar("_T")


class partial(_partial[_T]):  # noqa: N801
    """Partial which accepts Ellipsis for positional arguments."""

    @override
    def __call__(self, *args: Any, **kwargs: Any) -> _T:
        iter_args = iter(args)
        head = (next(iter_args) if arg is ... else arg for arg in self.args)
        return self.func(*head, *iter_args, **{**self.keywords, **kwargs})


__all__ = ["partial"]



from __future__ import annotations

import datetime as dt
from collections.abc import Callable, Iterable
from pathlib import Path
from re import search
from typing import Any

from typed_settings import default_loaders
from typed_settings.converters import TSConverter
from typed_settings.loaders import Loader
from typed_settings.types import AUTO, _Auto

from utilities.datetime import ensure_date, ensure_time, ensure_timedelta
from utilities.git import get_repo_root_or_cwd_sub_path
from utilities.pathlib import PathLike


def _config_toml(root: Path, /) -> Path | None:
    return path if (path := root.joinpath("config.toml")).exists() else None


CONFIG_FILES = [
    p for p in [get_repo_root_or_cwd_sub_path(_config_toml)] if p is not None
]


class ExtendedTSConverter(TSConverter):
    """An extension of the TSConverter for custom types."""

    def __init__(
        self,
        *,
        resolve_paths: bool = True,
        strlist_sep: str | Callable[[str], list] | None = ":",
    ) -> None:
        super().__init__(resolve_paths=resolve_paths, strlist_sep=strlist_sep)
        cases: list[tuple[type[Any], Callable[..., Any]]] = [
            (dt.date, ensure_date),
            (dt.time, ensure_time),
            (dt.timedelta, ensure_timedelta),
        ]
        try:
            from sqlalchemy import Engine

            from utilities.sqlalchemy import ensure_engine
        except ModuleNotFoundError:  # pragma: no cover
            pass
        else:
            cases.append((Engine, ensure_engine))

        def lift(
            cls: type[Any], func: Callable[[Any], Any], /
        ) -> Callable[[Any, type[Any]], Any]:
            def hook(value: Any, _: type[Any] = Any, /) -> Any:
                if not isinstance(value, cls | str):
                    msg = f"Could not convert value to {cls.__name__}: {value}"
                    raise TypeError(msg)
                return func(value)

            return hook

        extras = {cls: lift(cls, func) for cls, func in cases}
        self.scalar_converters |= extras


def get_loaders(
    *,
    appname: str = "appname",
    config_files: Iterable[PathLike] = CONFIG_FILES,
    config_file_section: str | _Auto = AUTO,
    config_files_var: None | str | _Auto = AUTO,
    env_prefix: None | str | _Auto = AUTO,
) -> list[Loader]:
    if search("_", appname):
        msg = f"{appname=}"
        raise GetLoadersError(msg)
    return default_loaders(
        appname,
        config_files=config_files,
        config_file_section=config_file_section,
        config_files_var=config_files_var,
        env_prefix=env_prefix,
    )


class GetLoadersError(Exception):
    ...


__all__ = ["CONFIG_FILES", "ExtendedTSConverter", "get_loaders", "GetLoadersError"]



from __future__ import annotations

import enum
import typing
from enum import auto
from operator import eq
from pathlib import Path
from typing import Any, TypedDict, cast

import sqlalchemy
from hypothesis import assume, given
from hypothesis.strategies import (
    DataObject,
    booleans,
    data,
    integers,
    none,
    permutations,
    sampled_from,
    sets,
    tuples,
)
from pytest import mark, param, raises
from sqlalchemy import (
    BIGINT,
    BINARY,
    BOOLEAN,
    CHAR,
    CLOB,
    DATE,
    DATETIME,
    DECIMAL,
    DOUBLE,
    DOUBLE_PRECISION,
    FLOAT,
    INT,
    INTEGER,
    NCHAR,
    NUMERIC,
    NVARCHAR,
    REAL,
    SMALLINT,
    TEXT,
    TIME,
    TIMESTAMP,
    UUID,
    VARBINARY,
    VARCHAR,
    BigInteger,
    Boolean,
    Column,
    Date,
    DateTime,
    Double,
    Engine,
    Float,
    Integer,
    Interval,
    LargeBinary,
    MetaData,
    Numeric,
    SmallInteger,
    String,
    Table,
    Text,
    Time,
    Unicode,
    UnicodeText,
    Uuid,
    select,
)
from sqlalchemy.exc import DatabaseError, NoSuchTableError
from sqlalchemy.orm import declarative_base

from utilities.hypothesis import (
    lists_fixed_length,
    sqlite_engines,
    temp_paths,
    text_ascii,
)
from utilities.sqlalchemy import (
    CheckColumnCollectionsEqualError,
    CheckColumnsEqualError,
    CheckColumnTypesBooleanEqualError,
    CheckColumnTypesDateTimeEqualError,
    CheckColumnTypesEnumEqualError,
    CheckColumnTypesEqualError,
    CheckColumnTypesFloatEqualError,
    CheckColumnTypesIntervalEqualError,
    CheckColumnTypesLargeBinaryEqualError,
    CheckColumnTypesNumericEqualError,
    CheckColumnTypesStringEqualError,
    CheckColumnTypesUuidEqualError,
    CheckEngineError,
    CheckTableOrColumnNamesEqualError,
    ParseEngineError,
    TableAlreadyExistsError,
    TablenameMixin,
    check_column_collections_equal,
    check_column_types_boolean_equal,
    check_column_types_datetime_equal,
    check_column_types_enum_equal,
    check_column_types_equal,
    check_column_types_float_equal,
    check_column_types_interval_equal,
    check_column_types_large_binary_equal,
    check_column_types_numeric_equal,
    check_column_types_string_equal,
    check_column_types_uuid_equal,
    check_columns_equal,
    check_dataframe_schema_against_table,
    check_engine,
    check_table_against_reflection,
    check_table_or_column_names_equal,
    check_tables_equal,
    columnwise_max,
    columnwise_min,
    create_engine,
    ensure_engine,
    ensure_tables_created,
    ensure_tables_dropped,
    get_table,
    get_table_name,
    insert_items,
    parse_engine,
    redirect_to_no_such_table_error,
    redirect_to_table_already_exists_error,
    reflect_table,
    serialize_engine,
)


class TestCheckColumnCollectionsEqual:
    def test_main(self) -> None:
        x = Table("x", MetaData(), Column("id", Integer, primary_key=True))
        check_column_collections_equal(x.columns, x.columns)

    def test_snake(self) -> None:
        x = Table("x", MetaData(), Column("id", Integer, primary_key=True))
        y = Table("y", MetaData(), Column("Id", Integer, primary_key=True))
        check_column_collections_equal(x.columns, y.columns, snake=True)

    def test_allow_permutations(self) -> None:
        x = Table(
            "x",
            MetaData(),
            Column("id1", Integer, primary_key=True),
            Column("id2", Integer, primary_key=True),
        )
        y = Table(
            "y",
            MetaData(),
            Column("id2", Integer, primary_key=True),
            Column("id1", Integer, primary_key=True),
        )
        check_column_collections_equal(x.columns, y.columns, allow_permutations=True)

    def test_snake_and_allow_permutations(self) -> None:
        x = Table(
            "x",
            MetaData(),
            Column("id1", Integer, primary_key=True),
            Column("id2", Integer, primary_key=True),
        )
        y = Table(
            "y",
            MetaData(),
            Column("Id2", Integer, primary_key=True),
            Column("Id1", Integer, primary_key=True),
        )
        check_column_collections_equal(
            x.columns, y.columns, snake=True, allow_permutations=True
        )

    @mark.parametrize(
        ("x", "y"),
        [
            param(
                Table("x", MetaData(), Column("id", Integer, primary_key=True)),
                Table(
                    "y",
                    MetaData(),
                    Column("id", Integer, primary_key=True),
                    Column("value", Integer),
                ),
            ),
            param(
                Table("x", MetaData(), Column("id1", Integer, primary_key=True)),
                Table("y", MetaData(), Column("id2", Integer, primary_key=True)),
            ),
        ],
    )
    def test_errors(self, *, x: Table, y: Table) -> None:
        with raises(CheckColumnCollectionsEqualError):
            check_column_collections_equal(x.columns, y.columns)


class TestCheckColumnsEqual:
    def test_equal(self) -> None:
        x = Column("id", Integer)
        check_columns_equal(x, x)

    def test_snake(self) -> None:
        x = Column("id", Integer)
        y = Column("Id", Integer)
        check_columns_equal(x, y, snake=True)

    def test_primary_key_off(self) -> None:
        x = Column("id", Integer, primary_key=True)
        y = Column("id", Integer, nullable=False)
        check_columns_equal(x, y, primary_key=False)

    @mark.parametrize(
        ("x", "y"),
        [
            param(Column("id", Integer, primary_key=True), Column("id", Integer)),
            param(Column("id", Integer), Column("id", Integer, nullable=False)),
        ],
    )
    def test_errors(self, *, x: Any, y: Any) -> None:
        with raises(CheckColumnsEqualError):
            check_columns_equal(x, y)


class TestCheckColumnTypesEqual:
    groups = (
        [BIGINT, INT, INTEGER, SMALLINT, BigInteger, Integer, SmallInteger],
        [BOOLEAN, Boolean],
        [DATE, Date],
        [DATETIME, TIMESTAMP, DateTime],
        [Interval],
        [BINARY, VARBINARY, LargeBinary],
        [
            DECIMAL,
            DOUBLE,
            DOUBLE_PRECISION,
            FLOAT,
            NUMERIC,
            REAL,
            Double,
            Float,
            Numeric,
        ],
        [
            CHAR,
            CLOB,
            NCHAR,
            NVARCHAR,
            TEXT,
            VARCHAR,
            String,
            Text,
            Unicode,
            UnicodeText,
            sqlalchemy.Enum,
        ],
        [TIME, Time],
        [UUID, Uuid],
    )

    @mark.parametrize(
        "cls",
        [
            param(Boolean),
            param(DateTime),
            param(Float),
            param(Interval),
            param(LargeBinary),
            param(Numeric),
            param(String),
            param(Unicode),
            param(UnicodeText),
            param(Uuid),
        ],
    )
    def test_equal_for_primaries(self, *, cls: type[Any]) -> None:
        check_column_types_equal(cls(), cls())

    def test_equal_for_primaries_enum(self) -> None:
        class Example(enum.Enum):
            member = auto()

        check_column_types_equal(sqlalchemy.Enum(Example), sqlalchemy.Enum(Example))

    @given(data=data())
    def test_equal_across_groups(self, *, data: DataObject) -> None:
        group = data.draw(sampled_from(self.groups))
        cls = data.draw(sampled_from(group))
        elements = sampled_from([cls, cls()])
        x, y = data.draw(lists_fixed_length(elements, 2))
        check_column_types_equal(x, y)

    @given(data=data())
    def test_unequal(self, *, data: DataObject) -> None:
        groups = self.groups
        i, j = data.draw(lists_fixed_length(integers(0, len(groups) - 1), 2))
        _ = assume(i != j)
        group_i, group_j = groups[i], groups[j]
        cls_x, cls_y = (data.draw(sampled_from(g)) for g in [group_i, group_j])
        x, y = (data.draw(sampled_from([c, c()])) for c in [cls_x, cls_y])
        with raises(CheckColumnTypesEqualError):
            check_column_types_equal(x, y)


class TestCheckColumnTypesBooleanEqual:
    @given(create_constraints=lists_fixed_length(booleans(), 2))
    def test_create_constraint(
        self, *, create_constraints: typing.Sequence[bool]
    ) -> None:
        create_constraint_x, create_constraint_y = create_constraints
        x, y = (Boolean(create_constraint=cs) for cs in create_constraints)
        if create_constraint_x is create_constraint_y:
            check_column_types_boolean_equal(x, y)
        else:
            with raises(CheckColumnTypesBooleanEqualError):
                check_column_types_boolean_equal(x, y)

    @given(names=lists_fixed_length(text_ascii(min_size=1) | none(), 2))
    def test_name(self, *, names: typing.Sequence[str | None]) -> None:
        name_x, name_y = names
        x, y = (Boolean(name=n) for n in names)
        if name_x == name_y:
            check_column_types_boolean_equal(x, y)
        else:
            with raises(CheckColumnTypesBooleanEqualError):
                check_column_types_boolean_equal(x, y)


class TestCheckColumnTypesDateTimeEqual:
    @given(timezones=lists_fixed_length(booleans(), 2))
    def test_main(self, *, timezones: typing.Sequence[bool]) -> None:
        timezone_x, timezone_y = timezones
        x, y = (DateTime(timezone=tz) for tz in timezones)
        if timezone_x is timezone_y:
            check_column_types_datetime_equal(x, y)
        else:
            with raises(CheckColumnTypesDateTimeEqualError):
                check_column_types_datetime_equal(x, y)


class TestCheckColumnTypesEnumEqual:
    def test_no_enum_classes(self) -> None:
        x = sqlalchemy.Enum()
        check_column_types_enum_equal(x, x)

    @given(data=data())
    def test_one_enum_class(self, *, data: DataObject) -> None:
        class Example(enum.Enum):
            member = auto()

        x = sqlalchemy.Enum(Example)
        y = sqlalchemy.Enum()
        x, y = data.draw(permutations([x, y]))
        with raises(CheckColumnTypesEnumEqualError):
            check_column_types_enum_equal(x, y)

    def test_two_enum_classes(self) -> None:
        class EnumX(enum.Enum):
            member = auto()

        class EnumY(enum.Enum):
            member = auto()

        x, y = (sqlalchemy.Enum(e) for e in [EnumX, EnumY])
        with raises(CheckColumnTypesEnumEqualError):
            check_column_types_enum_equal(x, y)

    @given(create_constraints=lists_fixed_length(booleans(), 2))
    def test_create_constraint(
        self, *, create_constraints: typing.Sequence[bool]
    ) -> None:
        class Example(enum.Enum):
            member = auto()

        create_constraint_x, create_constraint_y = create_constraints
        x, y = (
            sqlalchemy.Enum(Example, create_constraint=cs) for cs in create_constraints
        )
        if create_constraint_x is create_constraint_y:
            check_column_types_enum_equal(x, y)
        else:
            with raises(CheckColumnTypesEnumEqualError):
                check_column_types_enum_equal(x, y)

    @given(native_enums=lists_fixed_length(booleans(), 2))
    def test_native_enum(self, *, native_enums: typing.Sequence[bool]) -> None:
        class Example(enum.Enum):
            member = auto()

        native_enum_x, native_enum_y = native_enums
        x, y = (sqlalchemy.Enum(Example, native_enum=ne) for ne in native_enums)
        if native_enum_x is native_enum_y:
            check_column_types_enum_equal(x, y)
        else:
            with raises(CheckColumnTypesEnumEqualError):
                check_column_types_enum_equal(x, y)

    @given(lengths=lists_fixed_length(integers(6, 10), 2))
    def test_length(self, *, lengths: typing.Sequence[int]) -> None:
        class Example(enum.Enum):
            member = auto()

        length_x, length_y = lengths
        x, y = (sqlalchemy.Enum(Example, length=l_) for l_ in lengths)
        if length_x == length_y:
            check_column_types_enum_equal(x, y)
        else:
            with raises(CheckColumnTypesEnumEqualError):
                check_column_types_enum_equal(x, y)

    @given(inherit_schemas=lists_fixed_length(booleans(), 2))
    def test_inherit_schema(self, *, inherit_schemas: typing.Sequence[bool]) -> None:
        class Example(enum.Enum):
            member = auto()

        inherit_schema_x, inherit_schema_y = inherit_schemas
        x, y = (sqlalchemy.Enum(Example, inherit_schema=is_) for is_ in inherit_schemas)
        if inherit_schema_x is inherit_schema_y:
            check_column_types_enum_equal(x, y)
        else:
            with raises(CheckColumnTypesEnumEqualError):
                check_column_types_enum_equal(x, y)


class TestCheckColumnTypesFloatEqual:
    @given(precisions=lists_fixed_length(integers(0, 10) | none(), 2))
    def test_precision(self, *, precisions: typing.Sequence[int | None]) -> None:
        precision_x, precision_y = precisions
        x, y = (Float(precision=p) for p in precisions)
        if precision_x == precision_y:
            check_column_types_float_equal(x, y)
        else:
            with raises(CheckColumnTypesFloatEqualError):
                check_column_types_float_equal(x, y)

    @given(asdecimals=lists_fixed_length(booleans(), 2))
    def test_asdecimal(self, *, asdecimals: typing.Sequence[bool]) -> None:
        asdecimal_x, asdecimal_y = asdecimals
        x, y = (Float(asdecimal=cast(Any, a)) for a in asdecimals)
        if asdecimal_x is asdecimal_y:
            check_column_types_float_equal(x, y)
        else:
            with raises(CheckColumnTypesFloatEqualError):
                check_column_types_float_equal(x, y)

    @given(dec_ret_scales=lists_fixed_length(integers(0, 10) | none(), 2))
    def test_decimal_return_scale(
        self, *, dec_ret_scales: typing.Sequence[int | None]
    ) -> None:
        dec_ret_scale_x, dec_ret_scale_y = dec_ret_scales
        x, y = (Float(decimal_return_scale=drs) for drs in dec_ret_scales)
        if dec_ret_scale_x == dec_ret_scale_y:
            check_column_types_float_equal(x, y)
        else:
            with raises(CheckColumnTypesFloatEqualError):
                check_column_types_float_equal(x, y)


class TestCheckColumnTypesIntervalEqual:
    @given(natives=lists_fixed_length(booleans(), 2))
    def test_native(self, *, natives: typing.Sequence[bool]) -> None:
        native_x, native_y = natives
        x, y = (Interval(native=n) for n in natives)
        if native_x is native_y:
            check_column_types_interval_equal(x, y)
        else:
            with raises(CheckColumnTypesIntervalEqualError):
                check_column_types_interval_equal(x, y)

    @given(second_precisions=lists_fixed_length(integers(0, 10) | none(), 2))
    def test_second_precision(
        self, *, second_precisions: typing.Sequence[int | None]
    ) -> None:
        second_precision_x, second_precision_y = second_precisions
        x, y = (Interval(second_precision=sp) for sp in second_precisions)
        if second_precision_x == second_precision_y:
            check_column_types_interval_equal(x, y)
        else:
            with raises(CheckColumnTypesIntervalEqualError):
                check_column_types_interval_equal(x, y)

    @given(day_precisions=lists_fixed_length(integers(0, 10) | none(), 2))
    def test_day_precision(
        self, *, day_precisions: typing.Sequence[int | None]
    ) -> None:
        day_precision_x, day_precision_y = day_precisions
        x, y = (Interval(day_precision=dp) for dp in day_precisions)
        if day_precision_x == day_precision_y:
            check_column_types_interval_equal(x, y)
        else:
            with raises(CheckColumnTypesIntervalEqualError):
                check_column_types_interval_equal(x, y)


class TestCheckColumnTypesLargeBinaryEqual:
    @given(lengths=lists_fixed_length(integers(0, 10) | none(), 2))
    def test_main(self, *, lengths: typing.Sequence[int | None]) -> None:
        length_x, length_y = lengths
        x, y = (LargeBinary(length=l_) for l_ in lengths)
        if length_x == length_y:
            check_column_types_large_binary_equal(x, y)
        else:
            with raises(CheckColumnTypesLargeBinaryEqualError):
                check_column_types_large_binary_equal(x, y)


class TestCheckColumnTypesNumericEqual:
    @given(precisions=lists_fixed_length(integers(0, 10) | none(), 2))
    def test_precision(self, *, precisions: typing.Sequence[int | None]) -> None:
        precision_x, precision_y = precisions
        x, y = (Numeric(precision=p) for p in precisions)
        if precision_x == precision_y:
            check_column_types_numeric_equal(x, y)
        else:
            with raises(CheckColumnTypesNumericEqualError):
                check_column_types_numeric_equal(x, y)

    @given(asdecimals=lists_fixed_length(booleans(), 2))
    def test_asdecimal(self, *, asdecimals: typing.Sequence[bool]) -> None:
        asdecimal_x, asdecimal_y = asdecimals
        x, y = (Numeric(asdecimal=cast(Any, a)) for a in asdecimals)
        if asdecimal_x is asdecimal_y:
            check_column_types_numeric_equal(x, y)
        else:
            with raises(CheckColumnTypesNumericEqualError):
                check_column_types_numeric_equal(x, y)

    @given(scales=lists_fixed_length(integers(0, 10) | none(), 2))
    def test_numeric_scale(self, *, scales: typing.Sequence[int | None]) -> None:
        scale_x, scale_y = scales
        x, y = (Numeric(scale=s) for s in scales)
        if scale_x == scale_y:
            check_column_types_numeric_equal(x, y)
        else:
            with raises(CheckColumnTypesNumericEqualError):
                check_column_types_numeric_equal(x, y)

    @given(dec_ret_scales=lists_fixed_length(integers(0, 10) | none(), 2))
    def test_decimal_return_scale(
        self, *, dec_ret_scales: typing.Sequence[int | None]
    ) -> None:
        dec_ret_scale_x, dec_ret_scale_y = dec_ret_scales
        x, y = (Numeric(decimal_return_scale=drs) for drs in dec_ret_scales)
        if dec_ret_scale_x == dec_ret_scale_y:
            check_column_types_numeric_equal(x, y)
        else:
            with raises(CheckColumnTypesNumericEqualError):
                check_column_types_numeric_equal(x, y)


class TestCheckColumnTypesStringEqual:
    @given(
        cls=sampled_from([String, Unicode, UnicodeText]),
        lengths=lists_fixed_length(integers(0, 10) | none(), 2),
    )
    def test_length(
        self,
        *,
        cls: type[String | Unicode | UnicodeText],
        lengths: typing.Sequence[int | None],
    ) -> None:
        length_x, length_y = lengths
        x, y = (cls(length=l_) for l_ in lengths)
        if length_x == length_y:
            check_column_types_string_equal(x, y)
        else:
            with raises(CheckColumnTypesStringEqualError):
                check_column_types_string_equal(x, y)

    @given(collations=lists_fixed_length(text_ascii(min_size=1) | none(), 2))
    def test_collation(self, *, collations: typing.Sequence[str | None]) -> None:
        collation_x, collation_y = collations
        x, y = (String(collation=c) for c in collations)
        if collation_x == collation_y:
            check_column_types_string_equal(x, y)
        else:
            with raises(CheckColumnTypesStringEqualError):
                check_column_types_string_equal(x, y)


class TestCheckColumnTypesUuidEqual:
    @given(as_uuids=lists_fixed_length(booleans(), 2))
    def test_as_uuid(self, *, as_uuids: typing.Sequence[bool]) -> None:
        as_uuid_x, as_uuid_y = as_uuids
        x, y = (Uuid(as_uuid=cast(Any, au)) for au in as_uuids)
        if as_uuid_x is as_uuid_y:
            check_column_types_uuid_equal(x, y)
        else:
            with raises(CheckColumnTypesUuidEqualError):
                check_column_types_uuid_equal(x, y)

    @given(native_uuids=lists_fixed_length(booleans(), 2))
    def test_native_uuid(self, *, native_uuids: typing.Sequence[bool]) -> None:
        native_uuid_x, native_uuid_y = native_uuids
        x, y = (Uuid(native_uuid=nu) for nu in native_uuids)
        if native_uuid_x is native_uuid_y:
            check_column_types_uuid_equal(x, y)
        else:
            with raises(CheckColumnTypesUuidEqualError):
                check_column_types_uuid_equal(x, y)


class TestCheckDataFrameSchemaAgainstTable:
    def test_default(self) -> None:
        df_schema = {"a": int, "b": float}
        table = Table(
            "example",
            MetaData(),
            Column("id", Integer, primary_key=True),
            Column("a", Integer),
            Column("b", Float),
        )
        result = check_dataframe_schema_against_table(df_schema, table, eq)
        expected = {"a": "a", "b": "b"}
        assert result == expected

    def test_snake(self) -> None:
        df_schema = {"a": int, "b": float}
        table = Table(
            "example",
            MetaData(),
            Column("Id", Integer, primary_key=True),
            Column("A", Integer),
            Column("B", Float),
        )
        result = check_dataframe_schema_against_table(df_schema, table, eq, snake=True)
        expected = {"a": "A", "b": "B"}
        assert result == expected

    def test_df_has_extra_columns(self) -> None:
        df_schema = {"a": int, "b": float, "c": str}
        table = Table(
            "example",
            MetaData(),
            Column("id", Integer, primary_key=True),
            Column("a", Integer),
            Column("b", Float),
        )
        result = check_dataframe_schema_against_table(df_schema, table, eq)
        expected = {"a": "a", "b": "b"}
        assert result == expected

    def test_table_has_extra_columns(self) -> None:
        df_schema = {"a": int, "b": float}
        table = Table(
            "example",
            MetaData(),
            Column("id", Integer, primary_key=True),
            Column("a", Integer),
            Column("b", Float),
            Column("c", String),
        )
        result = check_dataframe_schema_against_table(df_schema, table, eq)
        expected = {"a": "a", "b": "b"}
        assert result == expected


class TestCheckEngine:
    @given(engine=sqlite_engines())
    def test_main(self, *, engine: Engine) -> None:
        check_engine(engine)

    @given(engine=sqlite_engines())
    def test_num_tables(self, *, engine: Engine) -> None:
        table = Table("example", MetaData(), Column("id", Integer, primary_key=True))
        ensure_tables_created(engine, table)
        check_engine(engine, num_tables=1)

    @given(engine=sqlite_engines())
    def test_num_tables_rel_tol(self, *, engine: Engine) -> None:
        table = Table("example", MetaData(), Column("id", Integer, primary_key=True))
        ensure_tables_created(engine, table)
        check_engine(engine, num_tables=2, rel_tol=0.5)

    @given(engine=sqlite_engines())
    def test_num_tables_abs_tol(self, *, engine: Engine) -> None:
        check_engine(engine, num_tables=1, abs_tol=1)

    @given(root=temp_paths())
    def test_error_unable_to_open(self, *, root: Path) -> None:
        engine = create_engine("sqlite", database=str(root))
        with raises(CheckEngineError):
            check_engine(engine)

    @given(engine=sqlite_engines())
    def test_error_num_tables_int(self, *, engine: Engine) -> None:
        with raises(CheckEngineError):
            check_engine(engine, num_tables=1)

    @given(engine=sqlite_engines())
    def test_error_num_tables_rel_tol(self, *, engine: Engine) -> None:
        with raises(CheckEngineError):
            check_engine(engine, num_tables=1, rel_tol=0.5)

    @given(engine=sqlite_engines())
    def test_error_num_tables_abs_tol(self, *, engine: Engine) -> None:
        with raises(CheckEngineError):
            check_engine(engine, num_tables=2, abs_tol=1)


class TestCheckTableAgainstReflection:
    @given(engine=sqlite_engines())
    def test_reflected(self, *, engine: Engine) -> None:
        table = Table("example", MetaData(), Column("Id", Integer, primary_key=True))
        ensure_tables_created(engine, table)
        check_table_against_reflection(table, engine)

    @given(engine=sqlite_engines())
    def test_error_no_such_table(self, *, engine: Engine) -> None:
        table = Table("example", MetaData(), Column("Id", Integer, primary_key=True))
        with raises(NoSuchTableError):
            _ = check_table_against_reflection(table, engine)


class TestCheckTablesEqual:
    def test_main(self) -> None:
        table = Table("example", MetaData(), Column("id", Integer, primary_key=True))
        check_tables_equal(table, table)

    def test_snake_table(self) -> None:
        x = Table("example", MetaData(), Column("id", Integer, primary_key=True))
        y = Table("Example", MetaData(), Column("id", Integer, primary_key=True))
        check_tables_equal(x, y, snake_table=True)

    def test_snake_columns(self) -> None:
        x = Table("example", MetaData(), Column("id", Integer, primary_key=True))
        y = Table("example", MetaData(), Column("Id", Integer, primary_key=True))
        check_tables_equal(x, y, snake_columns=True)

    def test_mapped_class(self) -> None:
        class Example(declarative_base()):
            __tablename__ = "example"

            Id = Column(Integer, primary_key=True)

        check_tables_equal(Example, Example)


class TestCheckTableOrColumnNamesEqual:
    @mark.parametrize(
        ("x", "y", "snake", "success"),
        [
            param("x", "x", False, True),
            param("x", "x", True, True),
            param("x", "X", False, False),
            param("x", "X", True, True),
            param("x", "y", False, False),
            param("x", "y", True, False),
        ],
    )
    def test_main(self, *, x: str, y: str, snake: bool, success: bool) -> None:
        if success:
            check_table_or_column_names_equal(x, y, snake=snake)
        else:
            with raises(CheckTableOrColumnNamesEqualError):
                check_table_or_column_names_equal(x, y, snake=snake)

    @mark.parametrize(("name", "expected"), [param(None, "Id"), param("x", "x")])
    def test_quoted_name(self, *, name: str | None, expected: str) -> None:
        class Kwargs(TypedDict, total=False):
            name: str

        class Example(declarative_base()):
            __tablename__ = "example"

            Id = Column(
                Integer,
                primary_key=True,
                **(cast(Kwargs, {} if name is None else {"name": name})),
            )

        check_table_or_column_names_equal(Example.Id.name, expected)


class TestColumnwiseMinMax:
    @given(
        engine=sqlite_engines(),
        values=sets(tuples(integers(0, 10) | none(), integers(0, 10) | none())),
    )
    def test_main(
        self, *, engine: Engine, values: set[tuple[int | None, int | None]]
    ) -> None:
        table = Table(
            "example",
            MetaData(),
            Column("id", Integer, primary_key=True, autoincrement=True),
            Column("x", Integer),
            Column("y", Integer),
        )
        ensure_tables_created(engine, table)
        insert_items(engine, ([{"x": x, "y": y} for x, y in values], table))
        sel = select(
            table.c["x"],
            table.c["y"],
            columnwise_min(table.c["x"], table.c["y"]).label("min_xy"),
            columnwise_max(table.c["x"], table.c["y"]).label("max_xy"),
        )
        with engine.begin() as conn:
            res = conn.execute(sel).all()
        assert len(res) == len(values)
        for x, y, min_xy, max_xy in res:
            if (x is None) and (y is None):
                assert min_xy is None
                assert max_xy is None
            elif (x is not None) and (y is None):
                assert min_xy == x
                assert max_xy == x
            elif (x is None) and (y is not None):
                assert min_xy == y
                assert max_xy == y
            else:
                assert min_xy == min(x, y)
                assert max_xy == max(x, y)

    @given(engine=sqlite_engines())
    def test_label(self, *, engine: Engine) -> None:
        table = Table(
            "example",
            MetaData(),
            Column("id_", Integer, primary_key=True, autoincrement=True),
            Column("x", Integer),
        )
        ensure_tables_created(engine, table)
        sel = select(columnwise_min(table.c.x, table.c.x))
        with engine.begin() as conn:
            _ = conn.execute(sel).all()


class TestCreateEngine:
    @given(temp_path=temp_paths())
    def test_main(self, *, temp_path: Path) -> None:
        engine = create_engine("sqlite", database=temp_path.name)
        assert isinstance(engine, Engine)

    @given(temp_path=temp_paths())
    def test_query(self, *, temp_path: Path) -> None:
        engine = create_engine(
            "sqlite",
            database=temp_path.name,
            query={"arg1": "value1", "arg2": ["value2"]},
        )
        assert isinstance(engine, Engine)


class TestEnsureEngine:
    @given(data=data(), engine=sqlite_engines())
    def test_main(self, *, data: DataObject, engine: Engine) -> None:
        maybe_engine = data.draw(
            sampled_from([engine, engine.url.render_as_string(hide_password=False)])
        )
        result = ensure_engine(maybe_engine)
        assert result.url == engine.url


class TestEnsureTablesCreated:
    @given(engine=sqlite_engines())
    @mark.parametrize("runs", [param(1), param(2)])
    def test_table(self, *, engine: Engine, runs: int) -> None:
        table = Table("example", MetaData(), Column("id_", Integer, primary_key=True))
        self._run_test(table, engine, runs)

    @given(engine=sqlite_engines())
    @mark.parametrize("runs", [param(1), param(2)])
    def test_mapped_class(self, *, engine: Engine, runs: int) -> None:
        class Example(declarative_base()):
            __tablename__ = "example"

            id_ = Column(Integer, primary_key=True)

        self._run_test(Example, engine, runs)

    def _run_test(
        self, table_or_mapped_class: Table | type[Any], engine: Engine, runs: int, /
    ) -> None:
        sel = get_table(table_or_mapped_class).select()
        with raises(NoSuchTableError), engine.begin() as conn:
            try:
                _ = conn.execute(sel).all()
            except DatabaseError as error:
                redirect_to_no_such_table_error(engine, error)
        for _ in range(runs):
            ensure_tables_created(engine, table_or_mapped_class)
        with engine.begin() as conn:
            _ = conn.execute(sel).all()


class TestEnsureTablesDropped:
    @given(engine=sqlite_engines())
    @mark.parametrize("runs", [param(1), param(2)])
    def test_table(self, *, engine: Engine, runs: int) -> None:
        table = Table("example", MetaData(), Column("id_", Integer, primary_key=True))
        self._run_test(table, engine, runs)

    @given(engine=sqlite_engines())
    @mark.parametrize("runs", [param(1), param(2)])
    def test_mapped_class(self, *, engine: Engine, runs: int) -> None:
        class Example(declarative_base()):
            __tablename__ = "example"

            id_ = Column(Integer, primary_key=True)

        self._run_test(Example, engine, runs)

    def _run_test(
        self, table_or_mapped_class: Table | type[Any], engine: Engine, runs: int, /
    ) -> None:
        table = get_table(table_or_mapped_class)
        sel = table.select()
        with engine.begin() as conn:
            table.create(conn)
            _ = conn.execute(sel).all()
        for _ in range(runs):
            ensure_tables_dropped(engine, table_or_mapped_class)
        with raises(NoSuchTableError), engine.begin() as conn:
            try:
                _ = conn.execute(sel).all()
            except DatabaseError as error:
                redirect_to_no_such_table_error(engine, error)


class TestGetTableName:
    def test_table(self) -> None:
        table = Table("example", MetaData(), Column("id_", Integer, primary_key=True))
        result = get_table_name(table)
        expected = "example"
        assert result == expected

    def test_mapped_class(self) -> None:
        class Example(declarative_base()):
            __tablename__ = "example"

            id_ = Column(Integer, primary_key=True)

        result = get_table_name(Example)
        expected = "example"
        assert result == expected


class TestParseEngine:
    @given(engine=sqlite_engines())
    def test_str(self, *, engine: Engine) -> None:
        url = engine.url
        result = parse_engine(url.render_as_string(hide_password=False))
        assert result.url == url

    def test_error(self) -> None:
        with raises(ParseEngineError):
            _ = parse_engine("error")


class TestRedirectToNoSuchSequenceError:
    @given(engine=sqlite_engines())
    def test_main(self, *, engine: Engine) -> None:
        seq = sqlalchemy.Sequence("example")
        with raises(NotImplementedError), engine.begin() as conn:
            _ = conn.scalar(seq)


class TestRedirectToNoSuchTableError:
    @given(engine=sqlite_engines())
    def test_main(self, *, engine: Engine) -> None:
        table = Table("example", MetaData(), Column("id", Integer, primary_key=True))
        with raises(NoSuchTableError), engine.begin() as conn:
            try:
                _ = conn.execute(select(table))
            except DatabaseError as error:
                redirect_to_no_such_table_error(engine, error)


class TestRedirectToTableAlreadyExistsError:
    @given(engine=sqlite_engines())
    def test_main(self, *, engine: Engine) -> None:
        table = Table("example", MetaData(), Column("id", Integer, primary_key=True))
        with engine.begin() as conn:
            _ = table.create(conn)
        with raises(TableAlreadyExistsError), engine.begin() as conn:
            try:
                _ = table.create(conn)
            except DatabaseError as error:
                redirect_to_table_already_exists_error(engine, error)


class TestReflectTable:
    @given(
        engine=sqlite_engines(),
        col_type=sampled_from(
            [
                INTEGER,
                INTEGER(),
                NVARCHAR,
                NVARCHAR(),
                NVARCHAR(1),
                Integer,
                Integer(),
                String,
                String(),
                String(1),
            ]
        ),
    )
    def test_main(self, *, engine: Engine, col_type: Any) -> None:
        table = Table("example", MetaData(), Column("Id", col_type, primary_key=True))
        ensure_tables_created(engine, table)
        reflected = reflect_table(table, engine)
        check_tables_equal(reflected, table)

    @given(engine=sqlite_engines())
    def test_error(self, *, engine: Engine) -> None:
        table = Table("example", MetaData(), Column("Id", Integer, primary_key=True))
        with raises(NoSuchTableError):
            _ = reflect_table(table, engine)


class TestSerializeEngine:
    @given(data=data())
    def test_main(self, *, data: DataObject) -> None:
        engine = data.draw(sqlite_engines())
        result = parse_engine(serialize_engine(engine))
        assert result.url == engine.url


class TestTablenameMixin:
    def test_main(self) -> None:
        class Example(declarative_base(cls=TablenameMixin)):
            Id = Column(Integer, primary_key=True)

        assert get_table_name(Example) == "example"



from __future__ import annotations

import re
from collections.abc import Callable
from typing import Any

from fastapi import APIRouter as _APIRouter
from fastapi.types import DecoratedCallable
from typing_extensions import override

_PATTERN = re.compile(r"(^/$)|(^.+[^\/]$)")


class APIRouter(_APIRouter):
    """Subclass which handles paths with & without trailing slashes."""

    @override
    def api_route(  # type: ignore
        self, *, path: str, include_in_schema: bool = True, **kwargs: Any
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        """N/A."""
        if _PATTERN.search(path):
            return super().api_route(
                path, include_in_schema=include_in_schema, **kwargs
            )
        msg = f"Invalid route: {path}"
        raise ValueError(msg)


__all__ = ["APIRouter"]



from __future__ import annotations

from collections import Counter
from collections.abc import Hashable, Iterable
from typing import Any, TypeGuard

from utilities.types import ensure_hashable


def check_duplicates(iterable: Iterable[Hashable], /) -> None:
    """Check if an iterable contains any duplicates."""
    dup = {k: v for k, v in Counter(iterable).items() if v > 1}
    if len(dup) >= 1:
        msg = f"{dup=}"
        raise CheckDuplicatesError(msg)


class CheckDuplicatesError(Exception):
    ...


def ensure_hashables(
    *args: Any, **kwargs: Any
) -> tuple[list[Hashable], dict[str, Hashable]]:
    """Ensure a set of positional & keyword arguments are all hashable."""
    hash_args = list(map(ensure_hashable, args))
    hash_kwargs = {k: ensure_hashable(v) for k, v in kwargs.items()}
    return hash_args, hash_kwargs


def is_iterable_not_str(obj: Any, /) -> TypeGuard[Iterable[Any]]:
    """Check if an object is iterable, but not a string."""
    try:
        iter(obj)
    except TypeError:
        return False
    return not isinstance(obj, str)


__all__ = ["check_duplicates", "CheckDuplicatesError", "is_iterable_not_str"]



from __future__ import annotations

from pathlib import Path

from hypothesis import given
from hypothesis.strategies import integers, none, sets
from sqlalchemy import Column, Engine, Integer, select
from sqlalchemy.orm import declarative_base

from utilities.fastparquet import get_dtypes
from utilities.hypothesis import sqlite_engines, temp_paths
from utilities.pandas import Int64
from utilities.sqlalchemy import ensure_tables_created, insert_items, select_to_parquet


class TestSelectToParquet:
    @given(
        engine=sqlite_engines(),
        ids=sets(integers(min_value=0, max_value=10), min_size=1, max_size=10),
        root=temp_paths(),
        stream=integers(1, 10) | none(),
    )
    def test_streamed_dataframe(
        self, *, engine: Engine, ids: set[int], root: Path, stream: int | None
    ) -> None:
        class Example(declarative_base()):  # does not work with a core table
            __tablename__ = "example"

            id_ = Column(Integer, primary_key=True)

        ensure_tables_created(engine, Example)
        insert_items(engine, ([(id_,) for id_ in ids], Example))
        sel = select(Example.id_)
        select_to_parquet(sel, engine, path := root.joinpath("df.parq"), stream=stream)
        dtypes = get_dtypes(path)
        assert dtypes == {"id_": Int64}



from __future__ import annotations

from collections.abc import Iterable, Iterator
from typing import cast

from bs4 import Tag
from polars import DataFrame

from utilities.text import ensure_str


def table_tag_to_dataframe(table: Tag, /) -> DataFrame:
    """Convert a `table` tag into a DataFrame."""

    def get_text(tag: Tag, child: str, /) -> list[str]:
        children = cast(Iterable[Tag], tag.find_all(child))
        return [ensure_str(x.string) for x in children]

    th_rows: list[str] | None = None
    td_rows: list[list[str]] = []
    for tr in cast(Iterable[Tag], table.find_all("tr")):
        if len(th := get_text(tr, "th")) >= 1:
            if th_rows is None:
                th_rows = th
            else:
                msg = f"{table=}"
                raise TableTagToDataFrameError(msg)
        if len(td := get_text(tr, "td")) >= 1:
            td_rows.append(td)
    cols = list(zip(*td_rows, strict=True))
    df = DataFrame(cols)
    if th_rows is None:
        return df
    return df.rename({f"column_{i}": th for i, th in enumerate(th_rows)})


class TableTagToDataFrameError(Exception):
    ...


def yield_tables(tag: Tag, /) -> Iterator[DataFrame]:
    return map(table_tag_to_dataframe, tag.find_all("table"))


__all__ = ["table_tag_to_dataframe", "TableTagToDataFrameError", "yield_tables"]



from __future__ import annotations

from datetime import tzinfo

from pytest import mark, param

from utilities.zoneinfo import HONG_KONG, TOKYO


class TestTimeZones:
    @mark.parametrize("timezone", [param(HONG_KONG), param(TOKYO)])
    def test_main(self, *, timezone: tzinfo) -> None:
        assert isinstance(timezone, tzinfo)



from __future__ import annotations

from pathlib import Path
from time import sleep

from utilities.cacher import cache_to_disk
from utilities.more_itertools import one


class TestCacheToDisk:
    def test_main(self, *, tmp_path: Path) -> None:
        path_inc = tmp_path.joinpath("increment")

        @cache_to_disk(root=tmp_path)
        def increment(x: int, /) -> int:
            return x + 1

        assert len(list(tmp_path.iterdir())) == 0
        assert increment(0) == 1
        assert set(tmp_path.iterdir()) == {path_inc}
        assert len(list(path_inc.iterdir())) == 1
        assert increment(0) == 1
        assert len(list(path_inc.iterdir())) == 1
        assert increment(1) == 2
        assert len(list(path_inc.iterdir())) == 2

    def test_ttl(self, *, tmp_path: Path) -> None:
        ttl = 0.1
        path_inc = tmp_path.joinpath("increment")

        @cache_to_disk(root=tmp_path, ttl=ttl)
        def increment(x: int, /) -> int:
            return x + 1

        assert len(list(tmp_path.iterdir())) == 0
        assert increment(0) == 1
        assert set(tmp_path.iterdir()) == {path_inc}
        path = one(path_inc.iterdir())
        orig = path.stat().st_mtime
        assert increment(0) == 1
        assert path.stat().st_mtime == orig
        sleep(2 * ttl)
        assert increment(0) == 1
        assert path.stat().st_mtime > orig

    def test_args_and_kwargs_resolved(self, *, tmp_path: Path) -> None:
        path_inc = tmp_path.joinpath("add")

        @cache_to_disk(root=tmp_path)
        def add(x: int, y: int) -> int:
            return x + y

        assert add(0, 0) == 0
        assert len(list(path_inc.iterdir())) == 1
        assert add(0, y=0) == 0
        assert len(list(path_inc.iterdir())) == 1
        assert add(x=0, y=0) == 0
        assert len(list(path_inc.iterdir())) == 1
        assert add(x=1, y=1) == 2
        assert len(list(path_inc.iterdir())) == 2



from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass
from functools import wraps
from typing import Any, Generic, TypeVar, cast

from memory_profiler import memory_usage
from typing_extensions import ParamSpec

_P = ParamSpec("_P")
_T = TypeVar("_T")


@dataclass(frozen=True)
class Output(Generic[_T]):
    """A function output, and its memory usage."""

    value: _T
    memory: float


def memory_profiled(func: Callable[_P, _T], /) -> Callable[_P, Output[_T]]:
    """Call a function, but also profile its maximum memory usage."""

    @wraps(func)
    def wrapped(*args: _P.args, **kwargs: _P.kwargs) -> Output[_T]:
        memory, value = memory_usage(
            cast(Any, (func, args, kwargs)), max_usage=True, retval=True
        )
        return Output(value=value, memory=memory)

    return wrapped


__all__ = ["memory_profiled", "Output"]



from __future__ import annotations

from collections.abc import Iterable
from re import escape
from typing import TypeVar

from more_itertools import one as _one

from utilities.errors import redirect_error

_T = TypeVar("_T")


def one(iterable: Iterable[_T], /) -> _T:
    """Return the only item from iterable."""
    try:
        return _one(iterable)
    except ValueError as error:
        (msg,) = error.args
        try:
            pattern = "too few items in iterable (expected 1)"
            redirect_error(error, escape(pattern), OneEmptyError(msg))
        except ValueError:
            pattern = (
                "Expected exactly one item in iterable, but got .*, .*, and "
                "perhaps more"
            )
            redirect_error(error, pattern, OneNonUniqueError(msg))


class OneError(Exception):
    ...


class OneEmptyError(OneError):
    ...


class OneNonUniqueError(OneError):
    ...


__all__ = ["one", "OneError", "OneEmptyError", "OneNonUniqueError"]



from __future__ import annotations

from utilities._xarray.common import (
    DataArray0,
    DataArray1,
    DataArray2,
    DataArray3,
    DataArrayB,
    DataArrayB0,
    DataArrayB1,
    DataArrayB2,
    DataArrayB3,
    DataArrayDns,
    DataArrayDns0,
    DataArrayDns1,
    DataArrayDns2,
    DataArrayDns3,
    DataArrayF,
    DataArrayF0,
    DataArrayF1,
    DataArrayF2,
    DataArrayF3,
    DataArrayI,
    DataArrayI0,
    DataArrayI1,
    DataArrayI2,
    DataArrayI3,
    DataArrayO,
    DataArrayO0,
    DataArrayO1,
    DataArrayO2,
    DataArrayO3,
)

__all__ = [
    "DataArray0",
    "DataArray1",
    "DataArray2",
    "DataArray3",
    "DataArrayB",
    "DataArrayB0",
    "DataArrayB1",
    "DataArrayB2",
    "DataArrayB3",
    "DataArrayDns",
    "DataArrayDns0",
    "DataArrayDns1",
    "DataArrayDns2",
    "DataArrayDns3",
    "DataArrayF",
    "DataArrayF0",
    "DataArrayF1",
    "DataArrayF2",
    "DataArrayF3",
    "DataArrayI",
    "DataArrayI0",
    "DataArrayI1",
    "DataArrayI2",
    "DataArrayI3",
    "DataArrayO",
    "DataArrayO0",
    "DataArrayO1",
    "DataArrayO2",
    "DataArrayO3",
]


try:
    from utilities._xarray.numbagg import ewma, exp_moving_sum
except (ImportError, ModuleNotFoundError):  # pragma: no cover
    pass
else:
    __all__ += ["ewma", "exp_moving_sum"]



from __future__ import annotations

import datetime as dt
from collections.abc import Hashable, Iterable, Iterator
from contextlib import contextmanager
from os import environ, getenv
from pathlib import Path
from re import search
from string import printable
from subprocess import run
from typing import TypedDict, cast

from hypothesis import HealthCheck, Phase, Verbosity, assume, settings
from hypothesis.errors import InvalidArgument
from hypothesis.strategies import (
    DrawFn,
    SearchStrategy,
    booleans,
    characters,
    composite,
    datetimes,
    integers,
    just,
    none,
)
from semver import Version

from utilities._hypothesis.common import (
    MaybeSearchStrategy,
    Shape,
    draw_text,
    floats_extra,
    lift_draw,
    lists_fixed_length,
    temp_dirs,
    temp_paths,
    text_ascii,
)
from utilities.datetime import UTC
from utilities.pathlib import temp_cwd
from utilities.text import ensure_str


@contextmanager
def assume_does_not_raise(
    *exceptions: type[Exception], match: str | None = None
) -> Iterator[None]:
    """Assume a set of exceptions are not raised.

    Optionally filter on the string representation of the exception.
    """
    try:
        yield
    except exceptions as caught:
        if match is None:
            _ = assume(condition=False)
        else:
            (msg,) = caught.args
            if search(match, ensure_str(msg)):
                _ = assume(condition=False)
            else:
                raise


@composite
def datetimes_utc(
    _draw: DrawFn,
    /,
    *,
    min_value: MaybeSearchStrategy[dt.datetime] = dt.datetime.min,
    max_value: MaybeSearchStrategy[dt.datetime] = dt.datetime.max,
) -> dt.datetime:
    """Strategy for generating datetimes with the UTC timezone."""
    draw = lift_draw(_draw)
    return draw(
        datetimes(
            min_value=draw(min_value).replace(tzinfo=None),
            max_value=draw(max_value).replace(tzinfo=None),
            timezones=just(UTC),
        )
    )


@composite
def git_repos(
    _draw: DrawFn, /, *, branch: MaybeSearchStrategy[str | None] = None
) -> Path:
    draw = lift_draw(_draw)
    path = draw(temp_paths())
    with temp_cwd(path):
        _ = run(["git", "init"], check=True)  # noqa: S603, S607
        _ = run(
            ["git", "config", "user.name", "User"],  # noqa: S603, S607
            check=True,
        )
        _ = run(
            ["git", "config", "user.email", "a@z.com"],  # noqa: S603, S607
            check=True,
        )
        file = path.joinpath("file")
        file.touch()
        file_str = str(file)
        _ = run(["git", "add", file_str], check=True)  # noqa: S603, S607
        _ = run(["git", "commit", "-m", "add"], check=True)  # noqa: S603, S607
        _ = run(["git", "rm", file_str], check=True)  # noqa: S603, S607
        _ = run(["git", "commit", "-m", "rm"], check=True)  # noqa: S603, S607
        if (branch := draw(branch)) is not None:
            _ = run(
                ["git", "checkout", "-b", branch],  # noqa: S603, S607
                check=True,
            )
    return path


def hashables() -> SearchStrategy[Hashable]:
    """Strategy for generating hashable elements."""
    return booleans() | integers() | none() | text_ascii()


_MAX_EXAMPLES: str = "MAX_EXAMPLES"
_NO_SHRINK: str = "NO_SHRINK"


def setup_hypothesis_profiles(
    *,
    max_examples: str = _MAX_EXAMPLES,
    no_shrink: str = _NO_SHRINK,
    suppress_health_check: Iterable[HealthCheck] = (),
) -> None:
    """Set up the hypothesis profiles."""

    def yield_phases() -> Iterator[Phase]:
        yield Phase.explicit
        yield Phase.reuse
        yield Phase.generate
        yield Phase.target
        if not bool(int(getenv(no_shrink, default="0"))):
            yield Phase.shrink

    phases = set(yield_phases())

    class Kwargs(TypedDict, total=False):
        verbosity: Verbosity

    for name, default_max_examples, verbosity in [
        ("dev", 10, None),
        ("default", 100, None),
        ("ci", 1000, None),
        ("debug", 10, Verbosity.verbose),
    ]:
        try:
            env_var = environ[max_examples]
        except KeyError:
            max_examples_use = default_max_examples
        else:
            max_examples_use = int(env_var)
        verbosity_use = cast(
            Kwargs, {} if verbosity is None else {"verbosity": verbosity}
        )
        settings.register_profile(
            name,
            max_examples=max_examples_use,
            **verbosity_use,
            phases=phases,
            report_multiple_bugs=True,
            suppress_health_check=suppress_health_check,
            deadline=None,
            print_blob=True,
        )
    settings.load_profile(getenv("HYPOTHESIS_PROFILE", "default"))


@composite
def slices(
    _draw: DrawFn,
    iter_len: int,
    /,
    *,
    slice_len: MaybeSearchStrategy[int | None] = None,
) -> slice:
    """Strategy for generating continuous slices from an iterable."""
    draw = lift_draw(_draw)
    if (slice_len_ := draw(slice_len)) is None:
        slice_len_ = draw(integers(0, iter_len))
    elif not 0 <= slice_len_ <= iter_len:
        msg = f"Slice length {slice_len_} exceeds iterable length {iter_len}"
        raise InvalidArgument(msg)
    start = draw(integers(0, iter_len - slice_len_))
    stop = start + slice_len_
    return slice(start, stop)


def text_clean(
    *,
    min_size: MaybeSearchStrategy[int] = 0,
    max_size: MaybeSearchStrategy[int | None] = None,
    disallow_na: MaybeSearchStrategy[bool] = False,
) -> SearchStrategy[str]:
    """Strategy for generating clean text."""
    return draw_text(
        characters(blacklist_categories=["Z", "C"]),
        min_size=min_size,
        max_size=max_size,
        disallow_na=disallow_na,
    )


def text_printable(
    *,
    min_size: MaybeSearchStrategy[int] = 0,
    max_size: MaybeSearchStrategy[int | None] = None,
    disallow_na: MaybeSearchStrategy[bool] = False,
) -> SearchStrategy[str]:
    """Strategy for generating printable text."""
    return draw_text(
        characters(whitelist_categories=[], whitelist_characters=printable),
        min_size=min_size,
        max_size=max_size,
        disallow_na=disallow_na,
    )


@composite
def versions(  # noqa: PLR0912
    _draw: DrawFn,
    /,
    *,
    min_version: MaybeSearchStrategy[Version | None] = None,
    max_version: MaybeSearchStrategy[Version | None] = None,
) -> Version:
    """Strategy for generating `Version`s."""
    draw = lift_draw(_draw)
    min_version_, max_version_ = (draw(mv) for mv in (min_version, max_version))
    if isinstance(min_version_, Version) and isinstance(max_version_, Version):
        if min_version_ > max_version_:
            msg = f"{min_version_=}, {max_version_=}"
            raise InvalidArgument(msg)
        major = draw(integers(min_version_.major, max_version_.major))
        minor, patch = draw(lists_fixed_length(integers(min_value=0), 2))
        version = Version(major=major, minor=minor, patch=patch)
        _ = assume(min_version_ <= version <= max_version_)
        return version
    if isinstance(min_version_, Version) and (max_version_ is None):
        major = draw(integers(min_value=min_version_.major))
        if major > min_version_.major:
            minor, patch = draw(lists_fixed_length(integers(min_value=0), 2))
        else:
            minor = draw(integers(min_version_.minor))
            if minor > min_version_.minor:
                patch = draw(integers(min_value=0))  # pragma: no cover
            else:
                patch = draw(integers(min_value=min_version_.patch))
    elif (min_version_ is None) and isinstance(max_version_, Version):
        major = draw(integers(0, max_version_.major))
        if major < max_version_.major:
            minor, patch = draw(lists_fixed_length(integers(min_value=0), 2))
        else:
            minor = draw(integers(0, max_version_.minor))
            if minor < max_version_.minor:
                patch = draw(integers(min_value=0))  # pragma: no cover
            else:
                patch = draw(integers(0, max_version_.patch))
    elif (min_version_ is None) and (max_version_ is None):
        major, minor, patch = draw(lists_fixed_length(integers(min_value=0), 3))
    else:
        msg = "Invalid case"  # pragma: no cover
        raise RuntimeError(msg)  # pragma: no cover
    return Version(major=major, minor=minor, patch=patch)


__all__ = [
    "assume_does_not_raise",
    "datetimes_utc",
    "draw_text",
    "git_repos",
    "floats_extra",
    "hashables",
    "lift_draw",
    "lists_fixed_length",
    "MaybeSearchStrategy",
    "setup_hypothesis_profiles",
    "Shape",
    "slices",
    "temp_dirs",
    "temp_paths",
    "text_ascii",
    "text_clean",
    "text_printable",
    "versions",
]


try:
    from utilities._hypothesis.luigi import namespace_mixins
except ModuleNotFoundError:  # pragma: no cover
    pass
else:
    __all__ += ["namespace_mixins"]


try:
    from utilities._hypothesis.numpy import (
        bool_arrays,
        concatenated_arrays,
        datetime64_arrays,
        datetime64_dtypes,
        datetime64_indexes,
        datetime64_kinds,
        datetime64_units,
        datetime64D_indexes,
        datetime64s,
        datetime64us_indexes,
        float_arrays,
        int32s,
        int64s,
        int_arrays,
        str_arrays,
        uint32s,
        uint64s,
    )
except ModuleNotFoundError:  # pragma: no cover
    pass
else:
    __all__ += [
        "bool_arrays",
        "concatenated_arrays",
        "datetime64_arrays",
        "datetime64_dtypes",
        "datetime64_indexes",
        "datetime64_kinds",
        "datetime64_units",
        "datetime64D_indexes",
        "datetime64s",
        "datetime64us_indexes",
        "float_arrays",
        "int_arrays",
        "int32s",
        "int64s",
        "str_arrays",
        "uint32s",
        "uint64s",
    ]


try:
    from utilities._hypothesis.pandas import (
        dates_pd,
        datetimes_pd,
        indexes,
        int_indexes,
        str_indexes,
        timestamps,
    )
except ModuleNotFoundError:  # pragma: no cover
    pass
else:
    __all__ += [
        "dates_pd",
        "datetimes_pd",
        "indexes",
        "int_indexes",
        "str_indexes",
        "timestamps",
    ]


try:
    from utilities._hypothesis.sqlalchemy import sqlite_engines
except ModuleNotFoundError:  # pragma: no cover
    pass
else:
    __all__ += ["sqlite_engines"]


try:
    from utilities._hypothesis.xarray import (
        bool_data_arrays,
        dicts_of_indexes,
        float_data_arrays,
        int_data_arrays,
        str_data_arrays,
    )
except ModuleNotFoundError:  # pragma: no cover
    pass
else:
    __all__ += [
        "bool_data_arrays",
        "dicts_of_indexes",
        "float_data_arrays",
        "int_data_arrays",
        "str_data_arrays",
    ]
